# coding=utf-8
from __future__ import unicode_literals

import os

import nose
from django.utils import six
from kgb import SpyAgency

from reviewboard.diffviewer.parser import DiffParserError
from reviewboard.scmtools.core import PRE_CREATION
from reviewboard.scmtools.errors import SCMError, FileNotFoundError
from reviewboard.scmtools.git import ShortSHA1Error, GitClient
from reviewboard.scmtools.models import Repository, Tool
from reviewboard.scmtools.tests.testcases import SCMTestCase


class GitTests(SpyAgency, SCMTestCase):
    """Unit tests for Git."""

    fixtures = ['test_scmtools']

    def setUp(self):
        super(GitTests, self).setUp()

        tool = Tool.objects.get(name='Git')

        self.local_repo_path = os.path.join(os.path.dirname(__file__),
                                            '..', 'testdata', 'git_repo')
        self.git_ssh_path = ('localhost:%s'
                             % self.local_repo_path.replace('\\', '/'))
        remote_repo_path = 'git@github.com:reviewboard/reviewboard.git'
        remote_repo_raw_url = ('http://github.com/api/v2/yaml/blob/show/'
                               'reviewboard/reviewboard/<revision>')

        self.repository = Repository(name='Git test repo',
                                     path=self.local_repo_path,
                                     tool=tool)
        self.remote_repository = Repository(name='Remote Git test repo',
                                            path=remote_repo_path,
                                            raw_file_url=remote_repo_raw_url,
                                            tool=tool)

        try:
            self.tool = self.repository.get_scmtool()
            self.remote_tool = self.remote_repository.get_scmtool()
        except ImportError:
            raise nose.SkipTest('git binary not found')

    def _read_fixture(self, filename):
        filename = os.path.join(os.path.dirname(__file__),
                                '..', 'testdata', filename)
        with open(filename, 'r') as f:
            return f.read()

    def _get_file_in_diff(self, diff, filenum=0):
        files = self.tool.get_parser(diff).parse()
        self.assertTrue(filenum < len(files))
        return files[filenum]

    def test_ssh(self):
        """Testing a SSH-backed git repository"""
        self._test_ssh(self.git_ssh_path)

    def test_ssh_with_site(self):
        """Testing a SSH-backed git repository with a LocalSite"""
        self._test_ssh_with_site(self.git_ssh_path)

    def test_filemode_diff(self):
        """Testing parsing filemode changes Git diff"""
        diff = self._read_fixture('git_filemode.diff')

        file = self._get_file_in_diff(diff)
        self.assertEqual(file.origFile, 'testing')
        self.assertEqual(file.newFile, 'testing')
        self.assertEqual(file.origInfo, 'e69de29')
        self.assertEqual(file.newInfo, 'bcae657')
        self.assertFalse(file.binary)
        self.assertFalse(file.deleted)
        self.assertFalse(file.is_symlink)
        self.assertEqual(file.data.splitlines()[0],
                         "diff --git a/testing b/testing")
        self.assertEqual(file.data.splitlines()[-1], "+ADD")
        self.assertEqual(file.insert_count, 1)
        self.assertEqual(file.delete_count, 0)

    def test_filemode_with_following_diff(self):
        """Testing parsing filemode changes with following Git diff"""
        diff = self._read_fixture('git_filemode2.diff')

        file = self._get_file_in_diff(diff)
        self.assertEqual(file.origFile, 'testing')
        self.assertEqual(file.newFile, 'testing')
        self.assertEqual(file.origInfo, 'e69de29')
        self.assertEqual(file.newInfo, 'bcae657')
        self.assertFalse(file.binary)
        self.assertFalse(file.deleted)
        self.assertFalse(file.is_symlink)
        self.assertEqual(file.data.splitlines()[0],
                         "diff --git a/testing b/testing")
        self.assertEqual(file.data.splitlines()[-1], "+ADD")
        self.assertEqual(file.insert_count, 1)
        self.assertEqual(file.delete_count, 0)

        file = self._get_file_in_diff(diff, 1)
        self.assertEqual(file.origFile, 'cfg/testcase.ini')
        self.assertEqual(file.newFile, 'cfg/testcase.ini')
        self.assertEqual(file.origInfo, 'cc18ec8')
        self.assertEqual(file.newInfo, '5e70b73')
        self.assertEqual(file.data.splitlines()[0],
                         "diff --git a/cfg/testcase.ini b/cfg/testcase.ini")
        self.assertEqual(file.data.splitlines()[-1], '+db = pyunit')
        self.assertEqual(file.insert_count, 2)
        self.assertEqual(file.delete_count, 1)

    def test_simple_diff(self):
        """Testing parsing simple Git diff"""
        diff = self._read_fixture('git_simple.diff')

        file = self._get_file_in_diff(diff)
        self.assertEqual(file.origFile, 'cfg/testcase.ini')
        self.assertEqual(file.newFile, 'cfg/testcase.ini')
        self.assertEqual(file.origInfo, 'cc18ec8')
        self.assertEqual(file.newInfo, '5e70b73')
        self.assertFalse(file.binary)
        self.assertFalse(file.deleted)
        self.assertFalse(file.is_symlink)
        self.assertEqual(len(file.data), 249)
        self.assertEqual(file.data.splitlines()[0],
                         "diff --git a/cfg/testcase.ini b/cfg/testcase.ini")
        self.assertEqual(file.data.splitlines()[-1], "+db = pyunit")
        self.assertEqual(file.insert_count, 2)
        self.assertEqual(file.delete_count, 1)

    def test_diff_with_unicode(self):
        """Testing parsing Git diff with unicode characters"""
        diff = ('diff --git a/cfg/téstcase.ini b/cfg/téstcase.ini\n'
                'index cc18ec8..5e70b73 100644\n'
                '--- a/cfg/téstcase.ini\n'
                '+++ b/cfg/téstcase.ini\n'
                '@@ -1,6 +1,7 @@\n'
                '+blah blah blah\n'
                ' [mysql]\n'
                ' hóst = localhost\n'
                ' pórt = 3306\n'
                ' user = user\n'
                ' pass = pass\n'
                '-db = pyunít\n'
                '+db = pyunít\n').encode('utf-8')

        file = self._get_file_in_diff(diff)
        self.assertEqual(file.origFile, 'cfg/téstcase.ini')
        self.assertEqual(file.newFile, 'cfg/téstcase.ini')
        self.assertEqual(file.origInfo, 'cc18ec8')
        self.assertEqual(file.newInfo, '5e70b73')
        self.assertFalse(file.binary)
        self.assertFalse(file.deleted)
        self.assertFalse(file.is_symlink)
        self.assertEqual(file.data.splitlines()[0].decode('utf-8'),
                         'diff --git a/cfg/téstcase.ini b/cfg/téstcase.ini')
        self.assertEqual(file.data.splitlines()[-1].decode('utf-8'),
                         '+db = pyunít')
        self.assertEqual(file.insert_count, 2)
        self.assertEqual(file.delete_count, 1)

    def test_diff_with_tabs_after_filename(self):
        """Testing parsing Git diffs with tabs after the filename"""
        diff = (
            b'diff --git a/README b/README\n'
            b'index 712544e4343bf04967eb5ea80257f6c64d6f42c7..'
            b'f88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1 100644\n'
            b'--- a/README\t\n'
            b'+++ b/README\t\n'
            b'@ -1,1 +1,1 @@\n'
            b'-blah blah\n'
            b'+blah\n'
            b'-\n'
            b'1.7.1\n')

        files = self.tool.get_parser(diff).parse()
        self.assertEqual(files[0].origFile, 'README')
        self.assertEqual(files[0].newFile, 'README')
        self.assertEqual(files[0].origInfo,
                         '712544e4343bf04967eb5ea80257f6c64d6f42c7')
        self.assertEqual(files[0].newInfo,
                         'f88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1')
        self.assertEqual(files[0].data, diff)
        self.assertEqual(files[0].insert_count, 1)
        self.assertEqual(files[0].delete_count, 2)

    def test_new_file_diff(self):
        """Testing parsing Git diff with new file"""
        diff = self._read_fixture('git_newfile.diff')

        file = self._get_file_in_diff(diff)
        self.assertEqual(file.origFile, 'IAMNEW')
        self.assertEqual(file.newFile, 'IAMNEW')
        self.assertEqual(file.origInfo, PRE_CREATION)
        self.assertEqual(file.newInfo, 'e69de29')
        self.assertFalse(file.binary)
        self.assertFalse(file.deleted)
        self.assertFalse(file.is_symlink)
        self.assertEqual(len(file.data), 123)
        self.assertEqual(file.data.splitlines()[0],
                         'diff --git a/IAMNEW b/IAMNEW')
        self.assertEqual(file.data.splitlines()[-1], '+Hello')
        self.assertEqual(file.insert_count, 1)
        self.assertEqual(file.delete_count, 0)

    def test_new_file_no_content_diff(self):
        """Testing parsing Git diff new file, no content"""
        diff = self._read_fixture('git_newfile_nocontent.diff')
        files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(files), 1)

        file = self._get_file_in_diff(diff)
        self.assertEqual(file.origFile, 'newfile')
        self.assertEqual(file.newFile, 'newfile')
        self.assertEqual(file.origInfo, PRE_CREATION)
        self.assertEqual(file.newInfo, 'e69de29')
        self.assertFalse(file.binary)
        self.assertFalse(file.deleted)
        self.assertFalse(file.is_symlink)
        lines = file.data.splitlines()
        self.assertEqual(len(lines), 3)
        self.assertEqual(lines[0], 'diff --git a/newfile b/newfile')
        self.assertEqual(file.insert_count, 0)
        self.assertEqual(file.delete_count, 0)

    def test_new_file_no_content_with_following_diff(self):
        """Testing parsing Git diff new file, no content, with following"""
        diff = self._read_fixture('git_newfile_nocontent2.diff')
        files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(files), 2)

        self.assertEqual(files[0].origFile, 'newfile')
        self.assertEqual(files[0].newFile, 'newfile')
        self.assertEqual(files[0].origInfo, PRE_CREATION)
        self.assertEqual(files[0].newInfo, 'e69de29')
        self.assertFalse(files[0].binary)
        self.assertFalse(files[0].deleted)
        self.assertFalse(files[0].is_symlink)
        lines = files[0].data.splitlines()
        self.assertEqual(len(lines), 3)
        self.assertEqual(lines[0], 'diff --git a/newfile b/newfile')
        self.assertEqual(files[0].insert_count, 0)
        self.assertEqual(files[0].delete_count, 0)

        self.assertEqual(files[1].origFile, 'cfg/testcase.ini')
        self.assertEqual(files[1].newFile, 'cfg/testcase.ini')
        self.assertEqual(files[1].origInfo, 'cc18ec8')
        self.assertEqual(files[1].newInfo, '5e70b73')
        lines = files[1].data.splitlines()
        self.assertEqual(len(lines), 13)
        self.assertEqual(lines[0],
                         'diff --git a/cfg/testcase.ini b/cfg/testcase.ini')
        self.assertEqual(lines[-1], '+db = pyunit')
        self.assertEqual(files[1].insert_count, 2)
        self.assertEqual(files[1].delete_count, 1)

    def test_del_file_diff(self):
        """Testing parsing Git diff with deleted file"""
        diff = self._read_fixture('git_delfile.diff')

        file = self._get_file_in_diff(diff)
        self.assertEqual(file.origFile, 'OLDFILE')
        self.assertEqual(file.newFile, 'OLDFILE')
        self.assertEqual(file.origInfo, '8ebcb01')
        self.assertEqual(file.newInfo, '0000000')
        self.assertFalse(file.binary)
        self.assertTrue(file.deleted)
        self.assertFalse(file.is_symlink)
        self.assertEqual(len(file.data), 132)
        self.assertEqual(file.data.splitlines()[0],
                         'diff --git a/OLDFILE b/OLDFILE')
        self.assertEqual(file.data.splitlines()[-1], '-Goodbye')
        self.assertEqual(file.insert_count, 0)
        self.assertEqual(file.delete_count, 1)

    def test_del_file_no_content_diff(self):
        """Testing parsing Git diff with deleted file, no content"""
        diff = (b'diff --git a/empty b/empty\n'
                b'deleted file mode 100644\n'
                b'index e69de29bb2d1d6434b8b29ae775ad8c2e48c5391..'
                b'0000000000000000000000000000000000000000\n')
        files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(files), 1)

        self.assertEqual(files[0].origFile, 'empty')
        self.assertEqual(files[0].newFile, 'empty')
        self.assertEqual(files[0].origInfo,
                         'e69de29bb2d1d6434b8b29ae775ad8c2e48c5391')
        self.assertEqual(files[0].newInfo,
                         '0000000000000000000000000000000000000000')
        self.assertFalse(files[0].binary)
        self.assertTrue(files[0].deleted)
        self.assertFalse(files[0].is_symlink)
        self.assertEqual(len(files[0].data), 141)
        self.assertEqual(files[0].data.splitlines()[0],
                         'diff --git a/empty b/empty')
        self.assertEqual(files[0].insert_count, 0)
        self.assertEqual(files[0].delete_count, 0)

    def test_del_file_no_content_with_following_diff(self):
        """Testing parsing Git diff with deleted file, no content, with
        following
        """
        diff = (b'diff --git a/empty b/empty\n'
                b'deleted file mode 100644\n'
                b'index e69de29bb2d1d6434b8b29ae775ad8c2e48c5391..'
                b'0000000000000000000000000000000000000000\n'
                b'diff --git a/foo/bar b/foo/bar\n'
                b'index 484ba93ef5b0aed5b72af8f4e9dc4cfd10ef1a81..'
                b'0ae4095ddfe7387d405bd53bd59bbb5d861114c5 100644\n'
                b'--- a/foo/bar\n'
                b'+++ b/foo/bar\n'
                b'@@ -1 +1,2 @@\n'
                b'+Hello!\n'
                b'blah\n')
        files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(files), 2)

        self.assertEqual(files[0].origFile, 'empty')
        self.assertEqual(files[0].newFile, 'empty')
        self.assertEqual(files[0].origInfo,
                         'e69de29bb2d1d6434b8b29ae775ad8c2e48c5391')
        self.assertEqual(files[0].newInfo,
                         '0000000000000000000000000000000000000000')
        self.assertFalse(files[0].binary)
        self.assertTrue(files[0].deleted)
        self.assertFalse(files[0].is_symlink)
        self.assertEqual(len(files[0].data), 141)
        self.assertEqual(files[0].data.splitlines()[0],
                         'diff --git a/empty b/empty')
        self.assertEqual(files[0].insert_count, 0)
        self.assertEqual(files[0].delete_count, 0)

        self.assertEqual(files[1].origFile, 'foo/bar')
        self.assertEqual(files[1].newFile, 'foo/bar')
        self.assertEqual(files[1].origInfo,
                         '484ba93ef5b0aed5b72af8f4e9dc4cfd10ef1a81')
        self.assertEqual(files[1].newInfo,
                         '0ae4095ddfe7387d405bd53bd59bbb5d861114c5')
        self.assertFalse(files[1].binary)
        self.assertFalse(files[1].deleted)
        self.assertFalse(files[1].is_symlink)
        lines = files[1].data.splitlines()
        self.assertEqual(len(lines), 7)
        self.assertEqual(lines[0], 'diff --git a/foo/bar b/foo/bar')
        self.assertEqual(lines[5], '+Hello!')
        self.assertEqual(files[1].insert_count, 1)
        self.assertEqual(files[1].delete_count, 0)

    def test_binary_diff(self):
        """Testing parsing Git diff with binary"""
        diff = self._read_fixture('git_binary.diff')

        file = self._get_file_in_diff(diff)
        self.assertEqual(file.origFile, 'pysvn-1.5.1.tar.gz')
        self.assertEqual(file.newFile, 'pysvn-1.5.1.tar.gz')
        self.assertEqual(file.origInfo, PRE_CREATION)
        self.assertEqual(file.newInfo, '86b520c')
        self.assertTrue(file.binary)
        self.assertFalse(file.deleted)
        self.assertFalse(file.is_symlink)
        lines = file.data.splitlines()
        self.assertEqual(len(lines), 4)
        self.assertEqual(
            lines[0], 'diff --git a/pysvn-1.5.1.tar.gz b/pysvn-1.5.1.tar.gz')
        self.assertEqual(
            lines[3], 'Binary files /dev/null and b/pysvn-1.5.1.tar.gz differ')
        self.assertEqual(file.insert_count, 0)
        self.assertEqual(file.delete_count, 0)

    def test_complex_diff(self):
        """Testing parsing Git diff with existing and new files"""
        diff = self._read_fixture('git_complex.diff')
        files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(files), 7)
        self.assertEqual(files[0].origFile, 'cfg/testcase.ini')
        self.assertEqual(files[0].newFile, 'cfg/testcase.ini')
        self.assertEqual(files[0].origInfo, '5e35098')
        self.assertEqual(files[0].newInfo, 'e254ef4')
        self.assertFalse(files[0].binary)
        self.assertFalse(files[0].deleted)
        self.assertFalse(files[0].is_symlink)
        self.assertEqual(files[0].insert_count, 2)
        self.assertEqual(files[0].delete_count, 1)
        self.assertEqual(len(files[0].data), 549)
        self.assertEqual(files[0].data.splitlines()[0],
                         'diff --git a/cfg/testcase.ini b/cfg/testcase.ini')
        self.assertEqual(files[0].data.splitlines()[13],
                         '         if isinstance(value, basestring):')

        self.assertEqual(files[1].origFile, 'tests/models.py')
        self.assertEqual(files[1].newFile, 'tests/models.py')
        self.assertEqual(files[1].origInfo, PRE_CREATION)
        self.assertEqual(files[1].newInfo, 'e69de29')
        self.assertFalse(files[1].binary)
        self.assertFalse(files[1].deleted)
        self.assertFalse(files[1].is_symlink)
        self.assertEqual(files[1].insert_count, 0)
        self.assertEqual(files[1].delete_count, 0)
        lines = files[1].data.splitlines()
        self.assertEqual(len(lines), 3)
        self.assertEqual(lines[0],
                         'diff --git a/tests/models.py b/tests/models.py')

        self.assertEqual(files[2].origFile, 'tests/tests.py')
        self.assertEqual(files[2].newFile, 'tests/tests.py')
        self.assertEqual(files[2].origInfo, PRE_CREATION)
        self.assertEqual(files[2].newInfo, 'e279a06')
        self.assertFalse(files[2].binary)
        self.assertFalse(files[2].deleted)
        self.assertFalse(files[2].is_symlink)
        self.assertEqual(files[2].insert_count, 2)
        self.assertEqual(files[2].delete_count, 0)
        lines = files[2].data.splitlines()
        self.assertEqual(len(lines), 8)
        self.assertEqual(lines[0],
                         'diff --git a/tests/tests.py b/tests/tests.py')
        self.assertEqual(lines[7],
                         '+This is some new content')

        self.assertEqual(files[3].origFile, 'pysvn-1.5.1.tar.gz')
        self.assertEqual(files[3].newFile, 'pysvn-1.5.1.tar.gz')
        self.assertEqual(files[3].origInfo, PRE_CREATION)
        self.assertEqual(files[3].newInfo, '86b520c')
        self.assertTrue(files[3].binary)
        self.assertFalse(files[3].deleted)
        self.assertFalse(files[3].is_symlink)
        self.assertEqual(files[3].insert_count, 0)
        self.assertEqual(files[3].delete_count, 0)
        lines = files[3].data.splitlines()
        self.assertEqual(len(lines), 4)
        self.assertEqual(
            lines[0], 'diff --git a/pysvn-1.5.1.tar.gz b/pysvn-1.5.1.tar.gz')
        self.assertEqual(lines[3],
                         'Binary files /dev/null and b/pysvn-1.5.1.tar.gz '
                         'differ')

        self.assertEqual(files[4].origFile, 'readme')
        self.assertEqual(files[4].newFile, 'readme')
        self.assertEqual(files[4].origInfo, '5e35098')
        self.assertEqual(files[4].newInfo, 'e254ef4')
        self.assertFalse(files[4].binary)
        self.assertFalse(files[4].deleted)
        self.assertFalse(files[4].is_symlink)
        self.assertEqual(files[4].insert_count, 1)
        self.assertEqual(files[4].delete_count, 1)
        lines = files[4].data.splitlines()
        self.assertEqual(len(lines), 7)
        self.assertEqual(lines[0], 'diff --git a/readme b/readme')
        self.assertEqual(lines[6], '+Hello there')

        self.assertEqual(files[5].origFile, 'OLDFILE')
        self.assertEqual(files[5].newFile, 'OLDFILE')
        self.assertEqual(files[5].origInfo, '8ebcb01')
        self.assertEqual(files[5].newInfo, '0000000')
        self.assertFalse(files[5].binary)
        self.assertTrue(files[5].deleted)
        self.assertFalse(files[5].is_symlink)
        self.assertEqual(files[5].insert_count, 0)
        self.assertEqual(files[5].delete_count, 1)
        lines = files[5].data.splitlines()
        self.assertEqual(len(lines), 7)
        self.assertEqual(lines[0], 'diff --git a/OLDFILE b/OLDFILE')
        self.assertEqual(lines[6], '-Goodbye')

        self.assertEqual(files[6].origFile, 'readme2')
        self.assertEqual(files[6].newFile, 'readme2')
        self.assertEqual(files[6].origInfo, '5e43098')
        self.assertEqual(files[6].newInfo, 'e248ef4')
        self.assertFalse(files[6].binary)
        self.assertFalse(files[6].deleted)
        self.assertFalse(files[6].is_symlink)
        self.assertEqual(files[6].insert_count, 1)
        self.assertEqual(files[6].delete_count, 1)
        lines = files[6].data.splitlines()
        self.assertEqual(len(lines), 7)
        self.assertEqual(lines[0], 'diff --git a/readme2 b/readme2')
        self.assertEqual(lines[6], '+Hello there')

    def test_parse_diff_with_index_range(self):
        """Testing Git diff parsing with an index range"""
        diff = (b'diff --git a/foo/bar b/foo/bar2\n'
                b'similarity index 88%\n'
                b'rename from foo/bar\n'
                b'rename to foo/bar2\n'
                b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
                b'e88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1 100644\n'
                b'--- a/foo/bar\n'
                b'+++ b/foo/bar2\n'
                b'@ -1,1 +1,1 @@\n'
                b'-blah blah\n'
                b'+blah\n')
        files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(files), 1)
        self.assertEqual(files[0].origFile, 'foo/bar')
        self.assertEqual(files[0].newFile, 'foo/bar2')
        self.assertEqual(files[0].origInfo,
                         '612544e4343bf04967eb5ea80257f6c64d6f42c7')
        self.assertEqual(files[0].newInfo,
                         'e88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1')
        self.assertEqual(files[0].insert_count, 1)
        self.assertEqual(files[0].delete_count, 1)

    def test_parse_diff_with_deleted_binary_files(self):
        """Testing Git diff parsing with deleted binary files"""
        diff = (b'diff --git a/foo.bin b/foo.bin\n'
                b'deleted file mode 100644\n'
                b'Binary file foo.bin has changed\n'
                b'diff --git a/bar.bin b/bar.bin\n'
                b'deleted file mode 100644\n'
                b'Binary file bar.bin has changed\n')
        files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(files), 2)
        self.assertEqual(files[0].origFile, 'foo.bin')
        self.assertEqual(files[0].newFile, 'foo.bin')
        self.assertEqual(files[0].binary, True)
        self.assertEqual(files[0].deleted, True)
        self.assertFalse(files[0].is_symlink)
        self.assertEqual(files[0].insert_count, 0)
        self.assertEqual(files[0].delete_count, 0)
        self.assertEqual(files[1].origFile, 'bar.bin')
        self.assertEqual(files[1].newFile, 'bar.bin')
        self.assertEqual(files[1].binary, True)
        self.assertEqual(files[1].deleted, True)
        self.assertFalse(files[1].is_symlink)
        self.assertEqual(files[1].insert_count, 0)
        self.assertEqual(files[1].delete_count, 0)

    def test_parse_diff_with_all_headers(self):
        """Testing Git diff parsing and preserving all headers"""
        preamble = (
            b'From 38d8fa94a9aa0c5b27943bec31d94e880165f1e0 Mon Sep '
            b'17 00:00:00 2001\n'
            b'From: Example Joe <joe@example.com>\n'
            b'Date: Thu, 5 Apr 2012 00:41:12 -0700\n'
            b'Subject: [PATCH 1/1] Sample patch.\n'
            b'\n'
            b'This is a test summary.\n'
            b'\n'
            b'With a description.\n'
            b'---\n'
            b' foo/bar |   2 -+n'
            b' README  |   2 -+n'
            b' 2 files changed, 2 insertions(+), 2 deletions(-)\n'
            b'\n')
        diff1 = (
            b'diff --git a/foo/bar b/foo/bar2\n'
            b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
            b'e88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1 100644\n'
            b'--- a/foo/bar\n'
            b'+++ b/foo/bar2\n'
            b'@ -1,1 +1,1 @@\n'
            b'-blah blah\n'
            b'+blah\n')
        diff2 = (
            b'diff --git a/README b/README\n'
            b'index 712544e4343bf04967eb5ea80257f6c64d6f42c7..'
            b'f88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1 100644\n'
            b'--- a/README\n'
            b'+++ b/README\n'
            b'@ -1,1 +1,1 @@\n'
            b'-blah blah\n'
            b'+blah\n'
            b'-\n'
            b'1.7.1\n')
        diff = preamble + diff1 + diff2

        files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(files), 2)
        self.assertEqual(files[0].origFile, 'foo/bar')
        self.assertEqual(files[0].newFile, 'foo/bar2')
        self.assertEqual(files[0].origInfo,
                         '612544e4343bf04967eb5ea80257f6c64d6f42c7')
        self.assertEqual(files[0].newInfo,
                         'e88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1')
        self.assertEqual(files[0].data, preamble + diff1)
        self.assertEqual(files[0].insert_count, 1)
        self.assertEqual(files[0].delete_count, 1)

        self.assertEqual(files[1].origFile, 'README')
        self.assertEqual(files[1].newFile, 'README')
        self.assertEqual(files[1].origInfo,
                         '712544e4343bf04967eb5ea80257f6c64d6f42c7')
        self.assertEqual(files[1].newInfo,
                         'f88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1')
        self.assertEqual(files[1].data, diff2)
        self.assertEqual(files[1].insert_count, 1)
        self.assertEqual(files[1].delete_count, 2)

    def test_parse_diff_revision(self):
        """Testing Git revision number parsing"""
        self.assertEqual(
            self.tool.parse_diff_revision('doc/readme', 'bf544ea'),
            ('doc/readme', 'bf544ea'))
        self.assertEqual(
            self.tool.parse_diff_revision('/dev/null', 'bf544ea'),
            ('/dev/null', PRE_CREATION))
        self.assertEqual(
            self.tool.parse_diff_revision('/dev/null', '0000000'),
            ('/dev/null', PRE_CREATION))

    def test_parse_diff_with_copy_and_rename_same_file(self):
        """Testing Git diff parsing with copy and rename of same file"""
        diff = (b'diff --git a/foo/bar b/foo/bar2\n'
                b'similarity index 100%\n'
                b'copy from foo/bar\n'
                b'copy to foo/bar2\n'
                b'diff --git a/foo/bar b/foo/bar3\n'
                b'similarity index 92%\n'
                b'rename from foo/bar\n'
                b'rename to foo/bar3\n'
                b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
                b'e88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1 100644\n'
                b'--- a/foo/bar\n'
                b'+++ b/foo/bar3\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-blah blah\n'
                b'+blah\n')
        files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(files), 2)

        f = files[0]
        self.assertEqual(f.origFile, 'foo/bar')
        self.assertEqual(f.newFile, 'foo/bar2')
        self.assertEqual(f.origInfo, '')
        self.assertEqual(f.newInfo, '')
        self.assertEqual(f.insert_count, 0)
        self.assertEqual(f.delete_count, 0)
        self.assertFalse(f.moved)
        self.assertTrue(f.copied)
        self.assertFalse(f.is_symlink)

        f = files[1]
        self.assertEqual(f.origFile, 'foo/bar')
        self.assertEqual(f.newFile, 'foo/bar3')
        self.assertEqual(f.origInfo,
                         '612544e4343bf04967eb5ea80257f6c64d6f42c7')
        self.assertEqual(f.newInfo,
                         'e88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1')
        self.assertEqual(f.insert_count, 1)
        self.assertEqual(f.delete_count, 1)
        self.assertTrue(f.moved)
        self.assertFalse(f.copied)
        self.assertFalse(f.is_symlink)

    def test_parse_diff_with_mode_change_and_rename(self):
        """Testing Git diff parsing with mode change and rename"""
        diff = (b'diff --git a/foo/bar b/foo/bar2\n'
                b'old mode 100755\n'
                b'new mode 100644\n'
                b'similarity index 99%\n'
                b'rename from foo/bar\n'
                b'rename to foo/bar2\n'
                b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
                b'e88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1\n'
                b'--- a/foo/bar\n'
                b'+++ b/foo/bar2\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-blah blah\n'
                b'+blah\n')
        files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(files), 1)

        f = files[0]
        self.assertEqual(f.origFile, 'foo/bar')
        self.assertEqual(f.newFile, 'foo/bar2')
        self.assertEqual(f.origInfo,
                         '612544e4343bf04967eb5ea80257f6c64d6f42c7')
        self.assertEqual(f.newInfo,
                         'e88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1')
        self.assertEqual(f.insert_count, 1)
        self.assertEqual(f.delete_count, 1)
        self.assertTrue(f.moved)
        self.assertFalse(f.copied)
        self.assertFalse(f.is_symlink)

    def test_diff_git_line_without_a_b(self):
        """Testing parsing Git diff with deleted file without a/ and
        b/ filename prefixes
        """
        diff = (b'diff --git foo foo\n'
                b'deleted file mode 100644\n'
                b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
                b'0000000000000000000000000000000000000000\n')

        files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(files), 1)

        f = files[0]
        self.assertEqual(f.origFile, 'foo')
        self.assertEqual(f.newFile, 'foo')
        self.assertTrue(f.deleted)
        self.assertFalse(f.is_symlink)

    def test_diff_git_line_without_a_b_quotes(self):
        """Testing parsing Git diff with deleted file without a/ and
        b/ filename prefixes and with quotes
        """
        diff = (b'diff --git "foo" "foo"\n'
                b'deleted file mode 100644\n'
                b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
                b'0000000000000000000000000000000000000000\n')

        files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(files), 1)

        f = files[0]
        self.assertEqual(f.origFile, 'foo')
        self.assertEqual(f.newFile, 'foo')
        self.assertTrue(f.deleted)
        self.assertFalse(f.is_symlink)

    def test_diff_git_line_without_a_b_and_spaces(self):
        """Testing parsing Git diff with deleted file without a/ and
        b/ filename prefixes and with spaces
        """
        diff = (b'diff --git foo bar1 foo bar1\n'
                b'deleted file mode 100644\n'
                b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
                b'0000000000000000000000000000000000000000\n')

        files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(files), 1)

        f = files[0]
        self.assertEqual(f.origFile, 'foo bar1')
        self.assertEqual(f.newFile, 'foo bar1')
        self.assertTrue(f.deleted)
        self.assertFalse(f.is_symlink)

    def test_diff_git_line_without_a_b_and_spaces_quotes(self):
        """Testing parsing Git diff with deleted file without a/ and
        b/ filename prefixes and with space and quotes
        """
        diff = (b'diff --git "foo bar1" "foo bar1"\n'
                b'deleted file mode 100644\n'
                b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
                b'0000000000000000000000000000000000000000\n')

        files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(files), 1)

        f = files[0]
        self.assertEqual(f.origFile, 'foo bar1')
        self.assertEqual(f.newFile, 'foo bar1')

    def test_diff_git_line_without_a_b_and_spaces_changed(self):
        """Testing parsing Git diff with deleted file without a/ and
        b/ filename prefixes and with spaces, with filename changes
        """
        diff = (b'diff --git foo bar1 foo bar2\n'
                b'deleted file mode 100644\n'
                b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
                b'0000000000000000000000000000000000000000\n')

        with self.assertRaises(DiffParserError) as cm:
            self.tool.get_parser(diff).parse()

        self.assertTrue(six.text_type(cm.exception).startswith(
            'Unable to parse the "diff --git" line'))

    def test_diff_git_line_without_a_b_and_spaces_quotes_changed(self):
        """Testing parsing Git diff with deleted file without a/ and
        b/ filename prefixes and with spaces and quotes, with filename
        changes
        """
        diff = (b'diff --git "foo bar1" "foo bar2"\n'
                b'deleted file mode 100644\n'
                b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
                b'0000000000000000000000000000000000000000\n'
                b'diff --git "foo bar1" foo\n'
                b'deleted file mode 100644\n'
                b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
                b'0000000000000000000000000000000000000000\n'
                b'diff --git foo "foo bar1"\n'
                b'deleted file mode 100644\n'
                b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
                b'0000000000000000000000000000000000000000\n')

        files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(files), 3)

        f = files[0]
        self.assertEqual(f.origFile, 'foo bar1')
        self.assertEqual(f.newFile, 'foo bar2')
        self.assertTrue(f.deleted)
        self.assertFalse(f.is_symlink)

        f = files[1]
        self.assertEqual(f.origFile, 'foo bar1')
        self.assertEqual(f.newFile, 'foo')

        f = files[2]
        self.assertEqual(f.origFile, 'foo')
        self.assertEqual(f.newFile, 'foo bar1')

    def test_diff_git_symlink_added(self):
        """Testing parsing Git diff with symlink added"""
        diff = (b'diff --git a/link b/link\n'
                b'new file mode 120000\n'
                b'index 0000000..100b938\n'
                b'--- /dev/null\n'
                b'+++ b/link\n'
                b'@@ -0,0 +1 @@\n'
                b'+README\n'
                b'\\ No newline at end of file\n')
        files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(files), 1)

        f = files[0]
        self.assertEqual(f.origInfo, PRE_CREATION)
        self.assertEqual(f.newFile, 'link')
        self.assertTrue(f.is_symlink)

    def test_diff_git_symlink_changed(self):
        """Testing parsing Git diff with symlink changed"""
        diff = (b'diff --git a/link b/link\n'
                b'index 100b937..100b938 120000\n'
                b'--- a/link\n'
                b'+++ b/link\n'
                b'@@ -1 +1 @@\n'
                b'-README\n'
                b'\\ No newline at end of file\n'
                b'+README.md\n'
                b'\\ No newline at end of file\n')
        files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(files), 1)

        f = files[0]
        self.assertEqual(f.newFile, 'link')
        self.assertEqual(f.origFile, 'link')
        self.assertTrue(f.is_symlink)

    def test_diff_git_symlink_removed(self):
        """Testing parsing Git diff with symlink removed"""
        diff = (b'diff --git a/link b/link\n'
                b'deleted file mode 120000\n'
                b'index 100b938..0000000\n'
                b'--- a/link\n'
                b'+++ /dev/null\n'
                b'@@ -1 +0,0 @@\n'
                b'-README.txt\n'
                b'\\ No newline at end of file\n')
        files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(files), 1)

        f = files[0]
        self.assertEqual(f.origFile, 'link')
        self.assertTrue(f.deleted)
        self.assertTrue(f.is_symlink)

    def test_file_exists(self):
        """Testing GitTool.file_exists"""
        self.assertTrue(self.tool.file_exists('readme', 'e965047'))
        self.assertTrue(self.tool.file_exists('readme', 'd6613f5'))

        self.assertTrue(not self.tool.file_exists('readme', PRE_CREATION))
        self.assertTrue(not self.tool.file_exists('readme', 'fffffff'))
        self.assertTrue(not self.tool.file_exists('readme2', 'fffffff'))

        # these sha's are valid, but commit and tree objects, not blobs
        self.assertTrue(not self.tool.file_exists('readme', 'a62df6c'))
        self.assertTrue(not self.tool.file_exists('readme2', 'ccffbb4'))

    def test_get_file(self):
        """Testing GitTool.get_file"""
        self.assertEqual(self.tool.get_file('readme', PRE_CREATION), b'')
        self.assertTrue(
            isinstance(self.tool.get_file('readme', 'e965047'), bytes))
        self.assertEqual(self.tool.get_file('readme', 'e965047'), b'Hello\n')
        self.assertEqual(self.tool.get_file('readme', 'd6613f5'),
                         b'Hello there\n')

        self.assertEqual(self.tool.get_file('readme'), b'Hello there\n')

        self.assertRaises(SCMError, lambda: self.tool.get_file(''))

        self.assertRaises(FileNotFoundError,
                          lambda: self.tool.get_file('', '0000000'))
        self.assertRaises(FileNotFoundError,
                          lambda: self.tool.get_file('hello', '0000000'))
        self.assertRaises(FileNotFoundError,
                          lambda: self.tool.get_file('readme', '0000000'))

    def test_parse_diff_revision_with_remote_and_short_SHA1_error(self):
        """Testing GitTool.parse_diff_revision with remote files and short
        SHA1 error
        """
        self.assertRaises(
            ShortSHA1Error,
            lambda: self.remote_tool.parse_diff_revision('README', 'd7e96b3'))

    def test_get_file_with_remote_and_short_SHA1_error(self):
        """Testing GitTool.get_file with remote files and short SHA1 error"""
        self.assertRaises(
            ShortSHA1Error,
            lambda: self.remote_tool.get_file('README', 'd7e96b3'))

    def test_valid_repository_https_username(self):
        """Testing GitClient.is_valid_repository with an HTTPS URL and external
        credentials
        """
        client = GitClient('https://example.com/test.git',
                           username='username',
                           password='pass/word')

        self.spy_on(client._run_git)
        client.is_valid_repository()

        self.assertEqual(client._run_git.calls[0].args[0],
                         ['ls-remote',
                          'https://username:pass%2Fword@example.com/test.git',
                          'HEAD'])

    def test_raw_file_url_error(self):
        """Testing Repository.get_file re-fetches when raw file URL changes"""
        self.spy_on(self.remote_repository._get_file_uncached,
                    call_fake=lambda a, b, x, y, z: 'first')
        self.assertEqual(self.remote_repository.get_file('PATH', 'd7e96b3'),
                         'first')
        # Ensure output of fake result matches.
        self.remote_repository._get_file_uncached.unspy()
        self.spy_on(self.remote_repository._get_file_uncached,
                    call_fake=lambda a, b, x, y, z: 'second')
        # Grab from cache when no changes and change fake result to confirm
        # it is not called.
        self.assertEqual(self.remote_repository.get_file('PATH', 'd7e96b3'),
                         'first')
        self.remote_repository.raw_file_url = (
            'http://github.com/api/v2/yaml/blob/show/reviewboard/<revision>')
        # When raw_file_url changed, do not grab from cache and ensure output
        # equals second fake value.
        self.assertEqual(self.remote_repository.get_file('PATH', 'd7e96b3'),
                         'second')

    def test_get_file_exists_caching_with_raw_url(self):
        """Testing Repository.get_file_exists properly checks file existence in
        repository or cache when raw file URL changes
        """
        self.spy_on(self.remote_repository._get_file_exists_uncached,
                    call_fake=lambda a, b, x, y, z: True)
        # Use spy to put key into cache
        self.assertTrue(self.remote_repository.get_file_exists('PATH',
                                                               'd7e96b3'))
        # Remove spy to ensure key is still in cache without needing spy
        self.remote_repository._get_file_exists_uncached.unspy()
        self.assertTrue(self.remote_repository.get_file_exists('PATH',
                                                               'd7e96b3'))
        self.remote_repository.raw_file_url = (
            'http://github.com/api/v2/yaml/blob/show/reviewboard/<revision>')
        # Does not exist when raw_file_url changed because it is not cached.
        self.assertFalse(self.remote_repository.get_file_exists('PATH',
                                                                'd7e96b3'))



"""Tests for the DiffContextResource."""

from __future__ import unicode_literals

from django.utils import six
from djblets.webapi.testing.decorators import webapi_test_template

from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import diff_context_mimetype
from reviewboard.webapi.tests.mixins import (BaseReviewRequestChildMixin,
                                             BasicTestsMetaclass)
from reviewboard.webapi.tests.urls import get_diff_context_url


@six.add_metaclass(BasicTestsMetaclass)
class ResourceTests(BaseWebAPITestCase, BaseReviewRequestChildMixin):
    """Testing the DiffContextResource APIs."""

    resource = resources.diff_context
    test_http_methods = ('GET',)

    sample_api_url = 'review-requests/<id>/diff-context/'

    fixtures = ['test_scmtools', 'test_users']

    def compare_item(self, item_rsp, obj):
        review_request, diffset, interdiffset = obj

        self.assertIn('revision', item_rsp)
        revision_info = item_rsp['revision']

        self.assertIn('interdiff_revision', revision_info)
        self.assertIn('is_draft_diff', revision_info)
        self.assertIn('is_draft_interdiff', revision_info)
        self.assertIn('is_interdiff', revision_info)
        self.assertIn('latest_revision', revision_info)
        self.assertIn('revision', revision_info)

        self.assertEqual(revision_info['is_draft_diff'], None)
        self.assertEqual(revision_info['is_draft_interdiff'], None)
        self.assertEqual(revision_info['is_interdiff'],
                         interdiffset is not None)
        self.assertEqual(revision_info['latest_revision'],
                         review_request.get_latest_diffset().revision)
        self.assertEqual(revision_info['revision'], diffset.revision)

        self.assertIn('num_diffs', item_rsp)

        if interdiffset:
            self.assertEqual(revision_info['interdiff_revision'],
                             interdiffset.revision)
            self.assertEqual(item_rsp['num_diffs'], 2)
        else:
            self.assertEqual(revision_info['interdiff_revision'],
                             None)
            self.assertEqual(item_rsp['num_diffs'], 1)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name,
                             with_interdiff=False):
        repository = self.create_repository(with_local_site=with_local_site,)
        review_request = self.create_review_request(with_local_site,
                                                    repository=repository,
                                                    public=True,
                                                    submitter=user)

        diffset = self.create_diffset(review_request=review_request,
                                      repository=repository)
        self.create_filediff(diffset)

        if with_interdiff:
            interdiffset = diffset

            diffset = self.create_diffset(review_request,
                                          repository=repository,
                                          revision=2)
            self.create_filediff(diffset)
        else:
            interdiffset = None

        return (
            get_diff_context_url(
                review_request_id=review_request.display_id,
                local_site_name=local_site_name),
            diff_context_mimetype,
            (review_request, diffset, interdiffset)
        )

    def setup_review_request_child_test(self, review_request):
        """Set up the review request child tests.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The test review request.

        Returns:
            tuple:
            A tuple of the API list URL and list mimetype to run tests on.
        """
        review_request.repository = self.create_repository()
        diffset = self.create_diffset(review_request)
        self.create_filediff(diffset)

        return (
            get_diff_context_url(review_request_id=review_request.display_id),
            diff_context_mimetype)

    @webapi_test_template
    def test_get_interdiff(self):
        """Testing the GET <URL> API with an interdiff"""
        url, mimetype, (review_request, diffset, interdiffset) = \
            self.setup_basic_get_test(user=self.user,
                                      with_local_site=False,
                                      local_site_name=None,
                                      with_interdiff=True)

        rsp = self.api_get(
            url,
            {
                'revision': diffset.revision,
                'interdiff-revision': interdiffset.revision,
            },
            expected_mimetype=mimetype)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn(self.resource.item_result_key, rsp)
        self.compare_item(rsp[self.resource.item_result_key],
                          (review_request, diffset, interdiffset))



from __future__ import unicode_literals

from django.core.exceptions import ObjectDoesNotExist
from djblets.util.decorators import augment_method_from
from djblets.webapi.decorators import (webapi_login_required,
                                       webapi_response_errors,
                                       webapi_request_fields)
from djblets.webapi.errors import (DOES_NOT_EXIST, INVALID_FORM_DATA,
                                   NOT_LOGGED_IN, PERMISSION_DENIED)

from reviewboard.webapi.decorators import webapi_check_local_site
from reviewboard.webapi.resources import resources
from reviewboard.webapi.resources.base_review_general_comment import \
    BaseReviewGeneralCommentResource


class ReviewGeneralCommentResource(BaseReviewGeneralCommentResource):
    """Provides information on general comments made on a review.

    If the review is a draft, then comments can be added, deleted, or
    changed on this list. However, if the review is already published,
    then no changes can be made.
    """
    allowed_methods = ('GET', 'POST', 'PUT', 'DELETE')
    policy_id = 'review_general_comment'
    model_parent_key = 'review'

    def get_queryset(self, request, review_id, *args, **kwargs):
        q = super(ReviewGeneralCommentResource, self).get_queryset(
            request, *args, **kwargs)
        return q.filter(review=review_id)

    @webapi_check_local_site
    @webapi_login_required
    @webapi_response_errors(DOES_NOT_EXIST, INVALID_FORM_DATA,
                            PERMISSION_DENIED, NOT_LOGGED_IN)
    @webapi_request_fields(
        required=BaseReviewGeneralCommentResource.REQUIRED_CREATE_FIELDS,
        optional=BaseReviewGeneralCommentResource.OPTIONAL_CREATE_FIELDS,
        allow_unknown=True
    )
    def create(self, request, *args, **kwargs):
        """Creates a general comment on a review.

        This will create a new comment on a review. The comment contains text
        only.

        Extra data can be stored later lookup. See
        :ref:`webapi2.0-extra-data` for more information.
        """
        try:
            review = resources.review.get_object(request, *args, **kwargs)
        except ObjectDoesNotExist:
            return DOES_NOT_EXIST

        if not resources.review.has_modify_permissions(request, review):
            return self.get_no_access_error(request.user)

        return self.create_comment(fields=(),
                                   review=review,
                                   comments_m2m=review.general_comments,
                                   **kwargs)

    @webapi_check_local_site
    @webapi_login_required
    @webapi_response_errors(DOES_NOT_EXIST, NOT_LOGGED_IN, PERMISSION_DENIED)
    @webapi_request_fields(
        optional=BaseReviewGeneralCommentResource.OPTIONAL_UPDATE_FIELDS,
        allow_unknown=True
    )
    def update(self, request, *args, **kwargs):
        """Updates a general comment.

        This can update the text or region of an existing comment. It
        can only be done for comments that are part of a draft review.

        Extra data can be stored later lookup. See
        :ref:`webapi2.0-extra-data` for more information.
        """
        try:
            resources.review_request.get_object(request, *args, **kwargs)
            review = resources.review.get_object(request, *args, **kwargs)
            general_comment = self.get_object(request, *args, **kwargs)
        except ObjectDoesNotExist:
            return DOES_NOT_EXIST

        return self.update_comment(request=request,
                                   review=review,
                                   comment=general_comment,
                                   **kwargs)

    @webapi_check_local_site
    @augment_method_from(BaseReviewGeneralCommentResource)
    def delete(self, *args, **kwargs):
        """Deletes the comment.

        This will remove the comment from the review. This cannot be undone.

        Only comments on draft reviews can be deleted. Attempting to delete
        a published comment will return a Permission Denied error.

        Instead of a payload response on success, this will return :http:`204`.
        """
        pass

    @webapi_check_local_site
    @augment_method_from(BaseReviewGeneralCommentResource)
    def get_list(self, *args, **kwargs):
        """Returns the list of general comments made on a review."""
        pass


review_general_comment_resource = ReviewGeneralCommentResource()



from __future__ import unicode_literals

import logging

from djblets.conditions import ConditionSet
from djblets.forms.fields import ConditionsField
from djblets.testing.decorators import add_fixtures
from kgb import SpyAgency

from reviewboard.integrations.forms import IntegrationConfigForm
from reviewboard.integrations.models import IntegrationConfig
from reviewboard.reviews.conditions import ReviewRequestConditionChoices
from reviewboard.testing.testcase import TestCase


class MyConfigForm(IntegrationConfigForm):
    my_conditions = ConditionsField(
        choices=ReviewRequestConditionChoices)


class IntegrationConfigTests(SpyAgency, TestCase):
    """Unit tests for reviewboard.integrations.models.IntegrationConfig."""

    def test_load_conditions(self):
        """Testing IntegrationConfig.load_conditions"""
        config = IntegrationConfig()
        config.settings['my_conditions'] = {
            'mode': 'all',
            'conditions': [
                {
                    'choice': 'branch',
                    'op': 'is',
                    'value': 'master',
                },
                {
                    'choice': 'summary',
                    'op': 'contains',
                    'value': '[WIP]',
                },
            ],
        }

        condition_set = config.load_conditions(MyConfigForm,
                                               conditions_key='my_conditions')

        self.assertEqual(condition_set.mode, ConditionSet.MODE_ALL)

        conditions = condition_set.conditions
        self.assertEqual(len(conditions), 2)

        condition = conditions[0]
        self.assertEqual(condition.choice.choice_id, 'branch')
        self.assertEqual(condition.operator.operator_id, 'is')
        self.assertEqual(condition.value, 'master')

        condition = conditions[1]
        self.assertEqual(condition.choice.choice_id, 'summary')
        self.assertEqual(condition.operator.operator_id, 'contains')
        self.assertEqual(condition.value, '[WIP]')

    def test_load_conditions_with_empty(self):
        """Testing IntegrationConfig.load_conditions with empty or missing
        data
        """
        config = IntegrationConfig()
        config.settings['conditions'] = None

        self.assertIsNone(config.load_conditions(MyConfigForm))

    def test_load_conditions_with_bad_data(self):
        """Testing IntegrationConfig.load_conditions with bad data"""
        config = IntegrationConfig()
        config.settings['conditions'] = 'dfsafas'

        self.spy_on(logging.debug)
        self.spy_on(logging.exception)

        self.assertIsNone(config.load_conditions(MyConfigForm))
        self.assertTrue(logging.debug.spy.called)
        self.assertTrue(logging.exception.spy.called)

    @add_fixtures(['test_users'])
    def test_match_conditions(self):
        """Testing IntegrationConfig.match_conditions"""
        config = IntegrationConfig()
        config.settings['my_conditions'] = {
            'mode': 'all',
            'conditions': [
                {
                    'choice': 'branch',
                    'op': 'is',
                    'value': 'master',
                },
                {
                    'choice': 'summary',
                    'op': 'contains',
                    'value': '[WIP]',
                },
            ],
        }

        review_request = self.create_review_request(
            branch='master',
            summary='[WIP] This is a test.')

        self.assertTrue(config.match_conditions(
            MyConfigForm,
            conditions_key='my_conditions',
            review_request=review_request))

        review_request = self.create_review_request(
            branch='master',
            summary='This is a test.')

        self.assertFalse(config.match_conditions(
            MyConfigForm,
            conditions_key='my_conditions',
            review_request=review_request))

    @add_fixtures(['test_users'])
    def test_match_conditions_sandbox(self):
        """Testing IntegrationConfig.match_conditions with exceptions
        sandboxed
        """
        config = IntegrationConfig()
        config.settings['my_conditions'] = {
            'mode': 'all',
            'conditions': [
                {
                    'choice': 'branch',
                    'op': 'is',
                    'value': 'master',
                },
                {
                    'choice': 'summary',
                    'op': 'contains',
                    'value': '[WIP]',
                },
            ],
        }

        self.create_review_request(
            branch='master',
            summary='[WIP] This is a test.')

        self.spy_on(logging.exception)

        self.assertFalse(config.match_conditions(
            MyConfigForm,
            conditions_key='my_conditions',
            review_request='test'))

        self.assertTrue(logging.exception.spy.called)



from __future__ import unicode_literals

import re
import warnings

from django.utils import six
from django.utils.translation import ugettext_lazy as _
from djblets.registries.registry import (ALREADY_REGISTERED,
                                         AlreadyRegisteredError,
                                         ATTRIBUTE_REGISTERED,
                                         DEFAULT_ERRORS,
                                         ItemLookupError,
                                         NOT_REGISTERED,
                                         Registry,
                                         UNREGISTER)
from djblets.urls.staticfiles import static_lazy
from djblets.util.decorators import augment_method_from

from reviewboard.deprecation import RemovedInReviewBoard40Warning


class TrophyType(object):
    """Base class for a type of trophy.

    Trophies are achievements that can be awarded to users based on some
    aspect of a review request. When a review request is filed, each registered
    trophy type (managed by the :py:data:`trophies` registry) will be checked
    using :py:meth:`qualifies` to see if the trophy can be awarded. If so, the
    trophy will be recorded and shown on the review request page.

    A trophy should include a displayable name, a category (essentially the
    ID of the trophy), and details for the trophy image.
    """

    #: The category of the trophy.
    #:
    #: This is the string ID of the trophy. For historical reasons, it's
    #: referred to as a category and not an ID.
    category = None

    #: The name of the trophy.
    name = None

    #: URLs for the trophy images.
    #:
    #: This is a dictionary of images, where each key is a resolution
    #: specifier (``1x``, ``2x``, etc.), and the value is a URL.
    #:
    #: Each must have widths/heights that are multipliers on the base
    #: width/height for the ``1x`` specifier.
    image_urls = {}

    #: The width of the base image.
    image_width = None

    #: The height of the base image.
    #:
    #: It is recommended to use a height of 48px max.
    image_height = None

    def __init__(self, title=None, image_url=None, image_width=None,
                 image_height=None):
        """Initialize the trophy.

        This accepts deprecated arguments for the trophy information.
        Subclasses should instead define the appropriate attributes on the
        class body.

        Args:
            title (unicode, optional):
                The name of the trophy.

                Subclasses should instead set :py:attr:`name`.

                .. deprecated:: 3.0

            image_url (unicode, optional):
                The URL of the trophy.

                Subclasses should instead set :py:attr:`image_urls`.

                .. deprecated:: 3.0

            image_width (int, optional):
                The width of the image.

                Subclasses should instead set :py:attr:`image_width`.

                .. deprecated:: 3.0

            image_height (int, optional):
                The height of the image.

                Subclasses should instead set :py:attr:`image_height`.

                .. deprecated:: 3.0
        """
        if not self.name and title:
            warnings.warn('%r should define "name" as a class attribute '
                          'instead of passing "title" to the constructor.'
                          % self.__class__,
                          RemovedInReviewBoard40Warning)

            self.name = title

        if not self.image_urls and image_url:
            warnings.warn('%r should define "image_urls" as a class attribute '
                          'instead of passing "image_url" to the constructor.'
                          % self.__class__,
                          RemovedInReviewBoard40Warning)

            self.image_urls = {
                '1x': image_url,
            }

        if not self.image_width:
            warnings.warn('%r should define "image_width" as a class '
                          'attribute instead of passing "image_width" to '
                          'the constructor.'
                          % self.__class__,
                          RemovedInReviewBoard40Warning)

            self.image_width = image_width or 32

        if not self.image_height:
            warnings.warn('%r should define "image_height" as a class '
                          'attribute instead of passing "image_height" to '
                          'the constructor.'
                          % self.__class__,
                          RemovedInReviewBoard40Warning)

            self.image_height = image_height or 48

    def get_display_text(self, trophy):
        """Return the text to display in the trophy banner.

        Args:
            trophy (reviewboard.accounts.models.Trophy):
                The stored trophy information.

        Returns:
            unicode:
            The display text for the trophy banner.
        """
        raise NotImplementedError

    def qualifies(self, review_request):
        """Return whether this trophy should be given to this review request.

        Args:
            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request to check for the trophy.

        Returns:
            bool:
            ``True`` if the trophy should be given, or ``False`` if not.
        """
        raise NotImplementedError

    def format_display_text(self, request, trophy, **kwargs):
        """Format the display text for the trophy.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            trophy (reviewboard.accounts.models.Trophy):
                The trophy instance.

            **kwargs (dict):
                Additional keyword arguments to use for formatting.

        Returns:
            unicode:
            The rendered text.
        """
        if getattr(self, 'display_format_str', None) is None:
            # In this case we may be dealing with a custom old-style trophy
            # that has implemented get_display_text. We need to check if
            # that has been implemented and, if so, use that.
            try:
                text = self.get_display_text(trophy)
            except NotImplementedError:
                raise NotImplementedError(
                    '%s does not define the format_display_text attribute.'
                    % type(self).__name__
                )
            else:
                warnings.warn(
                    'TrophyType.get_display_text has been deprecated in favor '
                    'of TrophyType.format_display_text.',
                    RemovedInReviewBoard40Warning)
                return text

        return self.display_format_str % dict(kwargs, **{
            'recipient': trophy.user.get_profile().get_display_name(
                getattr(request, 'user', None)),
            'review_request_id': trophy.review_request.display_id,
        })


class MilestoneTrophy(TrophyType):
    """A milestone trophy.

    It is awarded if review request ID is greater than 1000 and is a non-zero
    digit followed by only zeroes (e.g. 1000, 5000, 10000).
    """

    category = 'milestone'
    title = _('Milestone Trophy')

    image_urls = {
        '1x': static_lazy('rb/images/trophies/sparkly.png'),
        '2x': static_lazy('rb/images/trophies/sparkly@2x.png'),
    }

    image_width = 33
    image_height = 35

    display_format_str = _(
        '%(recipient)s got review request #%(review_request_id)d!'
    )

    def qualifies(self, review_request):
        """Return whether this trophy should be given to this review request.

        Args:
            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request to check for the trophy.

        Returns:
            bool:
            ``True`` if the trophy should be given, or ``False`` if not.
        """
        return (
            review_request.display_id >= 1000 and
            re.match(r'^[1-9]0+$', six.text_type(review_request.display_id))
        )


class FishTrophy(TrophyType):
    """A fish trophy.

    Give a man a fish, he'll waste hours trying to figure out why.
    """

    category = 'fish'
    name = _('Fish Trophy')

    image_urls = {
        '1x': static_lazy('rb/images/trophies/fish.png'),
        '2x': static_lazy('rb/images/trophies/fish@2x.png'),
    }

    image_width = 33
    image_height = 37

    display_format_str = _('%(recipient)s got a fish trophy!')

    def qualifies(self, review_request):
        """Return whether this trophy should be given to this review request.

        Args:
            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request to check for the trophy.

        Returns:
            bool:
            ``True`` if the trophy should be given, or ``False`` if not.
        """
        id_str = six.text_type(review_request.display_id)

        return (review_request.display_id >= 1000 and
                id_str == ''.join(reversed(id_str)))


class UnknownTrophy(TrophyType):
    """A trophy with an unknown category.

    The data for this trophy exists in the database but its category does not
    match the category of any registered trophy types.
    """

    name = 'Unknown Trophy'


class TrophyRegistry(Registry):
    lookup_attrs = ('category',)

    default_errors = dict(DEFAULT_ERRORS, **{
        ALREADY_REGISTERED: _(
            'Could not register trophy type %(item)s. This trophy type is '
            'already registered or its category conficts with another trophy.'
        ),
        ATTRIBUTE_REGISTERED: _(
            'Could not register trophy type %(item)s: Another trophy type '
            '(%(duplicate)s) is already registered with the same category.'
        ),
        NOT_REGISTERED: _(
            'No trophy type was found matching "%(attr_value)s".'
        ),
        UNREGISTER: _(
            'Could not unregister trophy type %(item)s: This trophy type '
            'was not yet registered.'
        ),
    })

    @augment_method_from(Registry)
    def register(self, trophy_type):
        """Register a new trophy type.

        Args:
            trophy_type (type):
                The trophy type (subclass of :py:class:`TrophyType`) to
                register.

        Raises:
            djblets.registries.errors.RegistrationError:
                The :py:attr:`TrophyType.category` value is missing on the
                trophy.

            djblets.registries.errors.AlreadyRegisteredError:
                This trophy type, or another with the same category, was
                already registered.
        """
        pass

    @augment_method_from(Registry)
    def unregister(self, trophy_type):
        """Unregister a trophy type.

        Args:
            trophy_type (type):
                The trophy type (subclass of :py:class:`TrophyType`) to
                unregister.

        Raises:
            djblets.registries.errors.ItemLookupError:
                This trophy type was not registered.
        """
        pass

    def get_for_category(self, category):
        """Return the TrophyType instance matching a given trophy category.

        If there's no registered trophy for the category,
        :py:class:`UnknownTrophy` will be returned.

        Args:
            category (unicode):
                The stored category for the trophy.

        Returns:
            TrophyType:
            The trophy matching the given category.
        """
        try:
            return self.get('category', category)
        except self.lookup_error_class:
            return UnknownTrophy

    def get_defaults(self):
        """Return the default trophies for the registry.

        This is used internally by the parent registry class to populate the
        list of default, buit-in trophies available to review requests.

        Returns:
            list of TrophyType:
            The list of default trophies.
        """
        return [
            MilestoneTrophy,
            FishTrophy,
        ]


#: The registry of available trophies.
trophies_registry = TrophyRegistry()


def register_trophy(trophy):
    """Register a TrophyType subclass.

    This will register a type of trophy. Review Board will use it to calculate
    and display possible trophies.

    Only :py:class:`TrophyType` subclasses are supported.

    .. deprecated:: 3.0

       This is deprecated in favor of calling ``trophies_registry.register``.

    Args:
        trophy (type):
            The trophy type (subclass of :py:class:`TrophyType`) to register.

    Raises:
        TypeError:
            The provided trophy to register is not a :py:class:`TrophyType`
            subclass.

        KeyError:
            The trophy could not be registered.
    """
    warnings.warn('register_trophy() is deprecated. Please use '
                  'reviewboard.accounts.trophies:trophies_registry.register() '
                  'instead.',
                  RemovedInReviewBoard40Warning)

    if not issubclass(trophy, TrophyType):
        raise TypeError('Only TrophyType subclasses can be registered')

    try:
        trophies_registry.register(trophy)
    except AlreadyRegisteredError as e:
        raise KeyError(six.text_type(e))


def unregister_trophy(trophy):
    """Unregister a TrophyType subclass.

    This will unregister a previously registered type of trophy.

    Only :py:class:`TrophyType` subclasses are supported. The class must have
    been registered beforehand or a :py:exc:`ValueError` will be thrown.

    .. deprecated:: 3.0

       This is deprecated in favor of calling ``trophies_registry.unregister``.

    Args:
        trophy (type):
            The trophy type (subclass of :py:class:`TrophyType`) to unregister.

    Raises:
        TypeError:
            The provided trophy to register is not a :py:class:`TrophyType`
            subclass.

        ValueError:
            The trophy could not be unregistered.
    """
    warnings.warn('unregister_trophy() is deprecated. Please use '
                  'reviewboard.accounts.trophies:trophies_registry'
                  '.unregister() instead.',
                  RemovedInReviewBoard40Warning)

    if not issubclass(trophy, TrophyType):
        raise TypeError('Only TrophyType subclasses can be unregistered')

    try:
        trophies_registry.unregister(trophy)
    except ItemLookupError as e:
        raise ValueError(six.text_type(e))


def get_registered_trophy_types():
    """Return all registered trophy types.

    .. deprecated:: 3.0

       This is deprecated in favor of iterating through
       :py:data:`trophies_registry`.

    Returns:
        list of TrophyType:
        The list of all registered trophies.
    """
    warnings.warn('get_registered_trophy_types() is deprecated. Please '
                  'iterate through '
                  'reviewboard.accounts.trophies:trophies_registry instead.',
                  RemovedInReviewBoard40Warning)

    return {
        trophy.category: trophy
        for trophy in trophies_registry
    }



"""Review Board feature checkers."""

from __future__ import unicode_literals

from djblets.features.checkers import SiteConfigFeatureChecker


class RBFeatureChecker(SiteConfigFeatureChecker):
    """Feature checker that checks against a LocalSite's configuration.

    Features can be enabled/disabled on a per-LocalSite basis by setting
    the specified feature ID to either ``True`` or ``False`` in the
    ``enabled_features`` key in that LocalSite's
    :py:attr:`~reviewboard.sites.models.LocalSite.extra_data`` field.

    If the key is absent, this checker will check against the site
    configuration (and then the Django settings) to see if it is enabled or
    disabled globally.
    """

    EXTRA_DATA_KEY = SiteConfigFeatureChecker.siteconfig_key

    def is_feature_enabled(self, feature_id, **kwargs):
        """Return whether a feature is enabled for a given ID.

        Features are strictly additive. That is, if a feature is enabled
        globally (e.g., via
        :py:class:`~djblets.siteconfig.models.SiteConfiguration` or via
        :file:`settings_local.py`), disabling it for a
        :py:class:`~reviewboard.site.models.LocalSite` will still result in the
        feature being available (i.e., this function will return ``True``).

        Args:
            feature_id (unicode):
                The unique identifier of the feature whose status is to be
                determined.

            **kwargs (dict):
                Additional keyword arguments.

        Keyword Args:
            request (django.http.HttpRequest):
                An optional request. If this request is made against a
                LocalSite, that LocalSite will be used to look up the feature.

                Either this argument or ``local_site`` must be provided to
                enable checking against a LocalSite.

                If provided, it will be used to cache the results of the
                :py:class:`~reviewboard.site.models.LocalSite` lookup.

            local_site (reviewboard.site.models.LocalSite):
                An optional local site. If provided, this LocalSite will be
                used to look up the status of the requested feature.

                Either this argument or ``request`` must be provided to enable
                checking against a LocalSite.

        Returns:
            bool:
            Whether or not the feature is enabled.
        """
        local_site = kwargs.get('local_site')
        request = kwargs.get('request')

        local_sites = []

        if local_site:
            local_sites.append(local_site)
        elif request is not None:
            if getattr(request, 'local_site', None):
                local_sites.append(request.local_site)

            if request.user.is_authenticated():
                if not hasattr(request, '_user_local_sites_cache'):
                    if request.user.is_authenticated():
                        local_sites.extend(request.user.local_site.all())

                    request._user_local_sites_cache = local_sites
                else:
                    local_sites = request._user_local_sites_cache

        for local_site in local_sites:
            if (local_site.extra_data and
                local_site.extra_data.get(self.EXTRA_DATA_KEY,
                                          {}).get(feature_id)):
                return True

        return super(RBFeatureChecker, self).is_feature_enabled(feature_id,
                                                                **kwargs)



from __future__ import unicode_literals

from reviewboard.hostingsvcs.service import HostingService


class FakeHostingService(HostingService):
    """A hosting service that is not provided by Review Board.

    Fake hosting services are intended to be used to advertise for Beanbag,
    Inc.'s Power Pack extension.
    """

    hosting_service_id = None


class FakeAWSCodeCommitHostingService(FakeHostingService):
    name = 'AWS CodeCommit'
    supported_scmtools = ['Git']
    hosting_service_id = 'aws-codecommit'


class FakeBitbucketServerHostingService(FakeHostingService):
    name = 'Bitbucket Server'
    supported_scmtools = ['Git']
    hosting_service_id = 'bitbucket-server'


class FakeGitHubEnterpriseHostingService(FakeHostingService):
    name = 'GitHub Enterprise'
    supported_scmtools = ['Git']
    hosting_service_id = 'github-enterprise'


class FakeVisualStudioTeamServicesHostingService(FakeHostingService):
    name = 'VisualStudio.com'
    supported_scmtools = [
        'Team Foundation Server',
        'Team Foundation Server (git)',
    ]
    hosting_service_id = 'visual-studio-online'


FAKE_HOSTING_SERVICES = {
    'rbpowerpack.hostingsvcs.aws_codecommit.AWSCodeCommit':
        FakeAWSCodeCommitHostingService,
    'rbpowerpack.hostingsvcs.bitbucket_server.BitbucketServer':
        FakeBitbucketServerHostingService,
    'rbpowerpack.hostingsvcs.githubenterprise.GitHubEnterprise':
        FakeGitHubEnterpriseHostingService,
    'rbpowerpack.hostingsvcs.visualstudio.VisualStudioTeamServices':
        FakeVisualStudioTeamServicesHostingService,
}



"""Tests for OAuth2 authentication with the web API."""

from __future__ import unicode_literals

from datetime import timedelta

from django.contrib.auth.models import User
from djblets.features.testing import override_feature_check
from djblets.testing.decorators import add_fixtures
from djblets.webapi.auth.backends import reset_auth_backends
from djblets.webapi.testing.testcases import WebAPITestCaseMixin

from reviewboard.admin.siteconfig import load_site_config
from reviewboard.oauth.features import oauth2_service_feature
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase
from reviewboard.webapi.tests.mimetypes import error_mimetype, session_mimetype
from reviewboard.webapi.tests.urls import get_session_url


class OAuth2TokenAuthTests(WebAPITestCaseMixin, TestCase):
    """Authentcaiton tests for OAuth2 tokens."""

    error_mimetype = error_mimetype
    fixtures = ['test_users']

    def setUp(self):
        super(OAuth2TokenAuthTests, self).setUp()

        self.owner = User.objects.get(username='doc')
        self.user = User.objects.get(username='grumpy')

    def tearDown(self):
        super(OAuth2TokenAuthTests, self).tearDown()

        load_site_config()
        reset_auth_backends()

    @classmethod
    def tearDownClass(cls):
        super(OAuth2TokenAuthTests, cls).tearDownClass()

        load_site_config()
        reset_auth_backends()

    def test_auth(self):
        """Testing OAuth2 authentication to the Web API with a valid token"""
        application = self.create_oauth_application(user=self.owner)
        token = self.create_oauth_token(application, self.user, 'session:read')

        with override_feature_check(oauth2_service_feature.feature_id, True):
            load_site_config()
            rsp = self.api_get(get_session_url(),
                               HTTP_AUTHORIZATION='Bearer %s' % token.token,
                               expected_mimetype=session_mimetype)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')

    def test_auth_disabled_app(self):
        """Testing OAuth2 authentication to the Web API with a valid token
        against a disabled app
        """
        application = self.create_oauth_application(user=self.owner,
                                                    enabled=False)
        token = self.create_oauth_token(application, self.user, 'session:read')

        with override_feature_check(oauth2_service_feature.feature_id, True):
            load_site_config()
            rsp = self.api_get(get_session_url(),
                               HTTP_AUTHORIZATION='Bearer %s' % token.token,
                               expected_status=401)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')

    def test_auth_feature_disabled(self):
        """Testing OAuth2 authentication to the Web API with a valid token
        with the feature disabled
        """
        application = self.create_oauth_application(user=self.owner)
        token = self.create_oauth_token(application, self.user, 'session:read')

        with override_feature_check(oauth2_service_feature.feature_id, False):
            load_site_config()
            rsp = self.api_get(get_session_url(),
                               HTTP_AUTHORIZATION='Bearer %s' % token.token,
                               expected_status=401)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')

    def test_auth_expired(self):
        """Testing OAuth2 authentication to the Web API with an expired token
        """
        application = self.create_oauth_application(user=self.owner)
        token = self.create_oauth_token(application, self.user, 'session:read',
                                        expires=timedelta(hours=-1))

        with override_feature_check(oauth2_service_feature.feature_id, True):
            load_site_config()
            rsp = self.api_get(get_session_url(),
                               HTTP_AUTHORIZATION='Bearer %s' % token.token,
                               expected_status=401)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')

    def test_auth_invalid_scope(self):
        """Testing OAuth2 authentication to the Web API with a token missing
        scopes"""
        application = self.create_oauth_application(user=self.owner)
        token = self.create_oauth_token(application, self.user)

        with override_feature_check(oauth2_service_feature.feature_id, True):
            load_site_config()
            rsp = self.api_get(get_session_url(),
                               HTTP_AUTHORIZATION='Bearer %s' % token.token,
                               expected_status=403)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')

    @add_fixtures(['test_site'])
    def test_auth_local_site(self):
        """Testing OAuth2 authentication to the Web API with a token limited to
        a Local Site
        """
        local_site = LocalSite.objects.get(pk=1)
        local_site.users.add(self.user)
        application = self.create_oauth_application(user=self.owner,
                                                    local_site=local_site)
        token = self.create_oauth_token(application, self.user, 'session:read')

        with override_feature_check(oauth2_service_feature.feature_id, True):
            load_site_config()
            rsp = self.api_get(get_session_url(),
                               HTTP_AUTHORIZATION='Bearer %s' % token.token,
                               expected_status=401)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')

    @add_fixtures(['test_site'])
    def test_auth_no_local_site(self):
        """Testing OAuth2 authentication to the Web API of a Local Site with an
        application not on that Local Site
        """
        local_site = LocalSite.objects.get(pk=1)
        local_site.users.add(self.user)
        application = self.create_oauth_application(user=self.owner)
        token = self.create_oauth_token(application, self.user, 'session:read')

        with override_feature_check(oauth2_service_feature.feature_id, True):
            load_site_config()
            rsp = self.api_get(get_session_url(local_site.name),
                               HTTP_AUTHORIZATION='Bearer %s' % token.token,
                               expected_status=401)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')

    @add_fixtures(['test_site'])
    def test_auth_no_local_site_access(self):
        """Testing OAuth2 authentication to the Web API of a Local Site with an
        application on that site without access to it
        """
        local_site = LocalSite.objects.get(pk=1)

        self.assertFalse(local_site.is_accessible_by(self.user))

        application = self.create_oauth_application(user=self.owner,
                                                    local_site=local_site)
        token = self.create_oauth_token(application, self.user, 'session:read')

        with override_feature_check(oauth2_service_feature.feature_id, True):
            load_site_config()
            rsp = self.api_get(get_session_url(local_site.name),
                               HTTP_AUTHORIZATION='Bearer %s' % token.token,
                               expected_status=401)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')

    @add_fixtures(['test_site'])
    def test_auth_local_site_public(self):
        """Testing OAuth2 authentication to the Web API of a public Local Site
        with an application on that Local Site
        """
        local_site = LocalSite.objects.get(pk=1)
        local_site.public = True
        local_site.save(update_fields=('public',))

        self.assertTrue(local_site.is_accessible_by(self.user))

        application = self.create_oauth_application(user=self.owner,
                                                    local_site=local_site)
        token = self.create_oauth_token(application, self.user, 'session:read')

        with override_feature_check(oauth2_service_feature.feature_id, True):
            load_site_config()
            rsp = self.api_get(get_session_url(local_site.name),
                               HTTP_AUTHORIZATION='Bearer %s' % token.token,
                               expected_mimetype=session_mimetype)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')

    @add_fixtures(['test_site'])
    def test_auth_local_site_member(self):
        """Testing OAuth2 authentication to the Web API of a Local Site with
        with an application on a that Local Site as a member
        """
        local_site = LocalSite.objects.get(pk=1)
        local_site.users.add(self.user)
        local_site.save(update_fields=('public',))

        self.assertTrue(local_site.is_accessible_by(self.user))

        application = self.create_oauth_application(user=self.owner,
                                                    local_site=local_site)
        token = self.create_oauth_token(application, self.user, 'session:read')

        with override_feature_check(oauth2_service_feature.feature_id, True):
            load_site_config()
            rsp = self.api_get(get_session_url(local_site.name),
                               HTTP_AUTHORIZATION='Bearer %s' % token.token,
                               expected_mimetype=session_mimetype)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')



from __future__ import unicode_literals

import logging
import os

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models import Max
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _
from djblets.db.fields import RelationCounterField

from reviewboard.admin.server import build_server_url
from reviewboard.attachments.managers import FileAttachmentManager
from reviewboard.attachments.mimetypes import MimetypeHandler
from reviewboard.diffviewer.models import FileDiff
from reviewboard.scmtools.models import Repository
from reviewboard.site.models import LocalSite


class FileAttachmentHistory(models.Model):
    """Revision history for a single file attachment.

    This tracks multiple revisions of the same file attachment (for instance,
    when someone replaces a screenshot with an updated version).
    """

    display_position = models.IntegerField()
    latest_revision = RelationCounterField('file_attachments')

    def get_revision_to_id_map(self):
        """Return a map from revision number to FileAttachment ID."""
        results = {}

        for attachment in self.file_attachments.all():
            results[attachment.attachment_revision] = attachment.id

        return results

    @staticmethod
    def compute_next_display_position(review_request):
        """Compute the display position for a new FileAttachmentHistory."""
        # Right now, display_position is monotonically increasing for each
        # review request. In the future this might be extended to allow the
        # user to change the order of attachments on the page.
        max_position = (
            FileAttachmentHistory.objects
            .filter(review_request=review_request)
            .aggregate(Max('display_position'))
            .get('display_position__max')) or 0

        return max_position + 1

    class Meta:
        db_table = 'attachments_fileattachmenthistory'
        verbose_name = _('File Attachment History')
        verbose_name_plural = _('File Attachment Histories')


@python_2_unicode_compatible
class FileAttachment(models.Model):
    """A file associated with a review request.

    Like diffs, a file can have comments associated with it.
    These comments are of type
    :py:class:`reviewboard.reviews.models.FileAttachmentComment`.
    """

    caption = models.CharField(_('caption'), max_length=256, blank=True)
    draft_caption = models.CharField(_('draft caption'),
                                     max_length=256, blank=True)
    orig_filename = models.CharField(_('original filename'),
                                     max_length=256, blank=True, null=True)
    user = models.ForeignKey(User,
                             blank=True,
                             null=True,
                             related_name='file_attachments')

    local_site = models.ForeignKey(LocalSite,
                                   blank=True,
                                   null=True,
                                   related_name='file_attachments')

    uuid = models.CharField(_('uuid'), max_length=255, blank=True)

    file = models.FileField(_('file'),
                            max_length=512,
                            blank=True,
                            null=True,
                            upload_to=os.path.join('uploaded', 'files',
                                                   '%Y', '%m', '%d'))
    mimetype = models.CharField(_('mimetype'), max_length=256, blank=True)

    # repo_path, repo_revision, and repository are used to identify
    # FileAttachments associated with committed binary files in a source tree.
    # They are not used for new files that don't yet have a revision.
    #
    # For new files, the added_in_filediff association is used.
    repo_path = models.CharField(_('repository file path'),
                                 max_length=1024,
                                 blank=True,
                                 null=True)
    repo_revision = models.CharField(_('repository file revision'),
                                     max_length=64,
                                     blank=True,
                                     null=True,
                                     db_index=True)
    repository = models.ForeignKey(Repository,
                                   blank=True,
                                   null=True,
                                   related_name='file_attachments')
    added_in_filediff = models.ForeignKey(FileDiff,
                                          blank=True,
                                          null=True,
                                          related_name='added_attachments')

    attachment_history = models.ForeignKey(FileAttachmentHistory,
                                           blank=True,
                                           null=True,
                                           related_name='file_attachments')
    attachment_revision = models.IntegerField(default=0)

    objects = FileAttachmentManager()

    @property
    def mimetype_handler(self):
        """Return the mimetype handler for this file."""
        if not hasattr(self, '_thumbnail'):
            self._thumbnail = MimetypeHandler.for_type(self)

        return self._thumbnail

    @property
    def review_ui(self):
        """Return the review UI for this file."""
        if not hasattr(self, '_review_ui'):
            from reviewboard.reviews.ui.base import FileAttachmentReviewUI
            self._review_ui = FileAttachmentReviewUI.for_type(self)

        return self._review_ui

    def _get_thumbnail(self):
        """Return the thumbnail for display."""
        if not self.mimetype_handler:
            return None

        try:
            return self.mimetype_handler.get_thumbnail()
        except Exception as e:
            logging.error('Error when calling get_thumbnail for '
                          'MimetypeHandler %r: %s',
                          self.mimetype_handler, e, exc_info=1)
            return None

    def _set_thumbnail(self, data):
        """Set the thumbnail."""
        if not self.mimetype_handler:
            return None

        try:
            self.mimetype_handler.set_thumbnail(data)
        except Exception as e:
            logging.error('Error when calling get_thumbnail for '
                          'MimetypeHandler %r: %s',
                          self.mimetype_handler, e, exc_info=1)
            return None

    thumbnail = property(_get_thumbnail, _set_thumbnail)

    @property
    def filename(self):
        """Return the filename for display purposes."""
        # Older versions of Review Board didn't store the original filename,
        # instead just using the FileField's name. Newer versions have
        # a dedicated filename field.
        if self.file:
            alt = os.path.basename(self.file.name)
        else:
            alt = None

        return self.orig_filename or alt

    @property
    def display_name(self):
        """Return a display name for the file."""
        if self.caption:
            return self.caption
        else:
            return self.filename

    @property
    def icon_url(self):
        """Return the icon URL for this file."""
        if not self.mimetype_handler:
            return None

        try:
            return self.mimetype_handler.get_icon_url()
        except Exception as e:
            logging.error('Error when calling get_thumbnail for '
                          'MimetypeHandler %r: %s',
                          self.mimetype_handler, e, exc_info=1)
            return None

    @property
    def is_from_diff(self):
        """Return if this file attachment is associated with a diff."""
        return (self.repository_id is not None or
                self.added_in_filediff_id is not None)

    @property
    def num_revisions(self):
        """Return the number of revisions of this attachment."""
        return FileAttachment.objects.filter(
            attachment_history=self.attachment_history_id).count() - 1

    def __str__(self):
        """Return a string representation of this file for the admin list."""
        return self.caption

    def get_review_request(self):
        """Return the ReviewRequest that this file is attached to."""
        if hasattr(self, '_review_request'):
            return self._review_request

        try:
            return self.review_request.all()[0]
        except IndexError:
            try:
                return self.inactive_review_request.all()[0]
            except IndexError:
                # Maybe it's on a draft.
                try:
                    draft = self.drafts.get()
                except ObjectDoesNotExist:
                    draft = self.inactive_drafts.get()

                return draft.review_request

    def get_comments(self):
        """Return all the comments made on this file attachment."""
        if not hasattr(self, '_comments'):
            self._comments = list(self.comments.all())

        return self._comments

    def get_absolute_url(self):
        """Return the absolute URL to download this file."""
        if not self.file:
            return None

        url = self.file.url

        if url.startswith('http:') or url.startswith('https:'):
            return url

        return build_server_url(url)

    def is_accessible_by(self, user):
        """Returns whether or not the user has access to this FileAttachment.

        This checks that the user has access to the LocalSite if the attachment
        is associated with a local site. This is only applicable for user owned
        file attachments.
        """
        return (self.user and user.is_authenticated() and
                (user.is_superuser or self.user == user) and
                (not self.local_site or
                 self.local_site.is_accessible_by(user)))

    def is_mutable_by(self, user):
        """Returns whether or not a user can modify this FileAttachment.

        This checks that the user is either a superuser or the owner of the
        file attachment. This is only applicable for user owned file
        attachments.
        """
        return self.user and (user.is_superuser or self.user == user)

    class Meta:
        db_table = 'attachments_fileattachment'
        get_latest_by = 'attachment_revision'
        verbose_name = _('File Attachment')
        verbose_name_plural = _('File Attachments')


def get_latest_file_attachments(file_attachments):
    """Filter the list of file attachments to only return the latest revisions.

    Args:
        file_attachments (list of
                          reviewboard.attachments.models.FileAttachment):
            The file attachments to filter.

    Returns:
        list of reviewboard.attachments.models.FileAttachment:
        The list of file attachments that are the latest revisions in their
        respective histories.
    """
    file_attachment_histories = FileAttachmentHistory.objects.filter(
        file_attachments__in=file_attachments)
    latest = {
        data['id']: data['latest_revision']
        for data in file_attachment_histories.values('id', 'latest_revision')
    }

    return [
        f
        for f in file_attachments
        if (not f.is_from_diff and
            f.attachment_revision == latest[f.attachment_history_id])
    ]



"""Feature definitions for OAuth2 integration."""

from __future__ import unicode_literals

from django.utils.translation import ugettext_lazy as _
from djblets.features.feature import Feature, FeatureLevel


class OAuth2ServiceFeature(Feature):
    """A feature for OAuth2 integration."""

    feature_id = 'oauth.service'
    name = _('OAuth2 Service Integration')
    level = FeatureLevel.STABLE
    summary = _('Allow Review Board to act as an OAuth2 authentication '
                'service for third-party apps.')


oauth2_service_feature = OAuth2ServiceFeature()



"""Internal support for handling deprecations in Review Board.

The version-specific objects in this module are not considered stable between
releases, and may be removed at any point. The base objects are considered
stable.
"""

from __future__ import unicode_literals


class BaseRemovedInReviewBoardVersionWarning(DeprecationWarning):
    """Base class for a Review Board deprecation warning.

    All version-specific deprecation warnings inherit from this, allowing
    callers to check for Review Board deprecations without being tied to a
    specific version.
    """


class RemovedInReviewBoard40Warning(BaseRemovedInReviewBoardVersionWarning):
    """Deprecations for features removed in Review Board 4.0.

    Note that this class will itself be removed in Review Board 4.0. If you
    need to check against Review Board deprecation warnings, please see
    :py:class:`BaseRemovedInReviewBoardVersionWarning`. Alternatively, you
    can use the alias for this class,
    :py:data:`RemovedInNextReviewBoardVersionWarning`.
    """


#: An alias for the next release of Djblets where features would be removed.
RemovedInNextReviewBoardVersionWarning = RemovedInReviewBoard40Warning



from __future__ import unicode_literals

from django.conf.urls import url
from django.contrib.auth import views as auth_views

from reviewboard.accounts.forms.auth import AuthenticationForm
from reviewboard.accounts import views as accounts_views


urlpatterns = [
    url(r'^login/$',
        auth_views.login,
        kwargs={
            'template_name': 'accounts/login.html',
            'authentication_form': AuthenticationForm,
        },
        name='login'),
    url(r'^logout/$',
        auth_views.logout_then_login,
        name='logout'),
    url(r'^preferences/$',
        accounts_views.MyAccountView.as_view(),
        name='user-preferences'),
    url(r'^preferences/preview-email/password-changed/'
        r'(?P<message_format>(text|html))/$',
        'preview_password_changed_email',
        name='preview-password-change-email'),
    url(r'^register/$',
        accounts_views.account_register,
        kwargs={
            'next_url': 'dashboard',
        },
        name='register'),
    url(r'^recover/$',
        auth_views.password_reset,
        kwargs={
            'template_name': 'accounts/password_reset.html',
            'email_template_name': 'accounts/password_reset_email.txt'
        },
        name='recover'),
    url(r'^recover/done/$',
        auth_views.password_reset_done,
        kwargs={
            'template_name': 'accounts/password_reset_done.html',
        },
        name='password_reset_done'),
    url(r'^reset/(?P<uidb64>[0-9A-Za-z_\-]+)/'
        r'(?P<token>[0-9A-Za-z]{1,13}-[0-9A-Za-z]{1,20})/$',
        auth_views.password_reset_confirm,
        kwargs={
            'template_name': 'accounts/password_reset_confirm.html',
        },
        name='password_reset_confirm'),
    url(r'^reset/done/$',
        auth_views.password_reset_complete,
        kwargs={
            'template_name': 'accounts/password_reset_complete.html',
        },
        name='password_reset_complete'),
    url(r'^preferences/oauth2-application/(?:(?P<app_id>[0-9]+)/)?$',
        accounts_views.edit_oauth_app,
        name='edit-oauth-app'),
    url(r'^preferences/preview-email/password-changed/$',
        accounts_views.preview_password_changed_email,
        name='preview-password-change-email'),
]



"""OAuth2 views."""

from __future__ import unicode_literals

from oauth2_provider.views import AuthorizationView as BaseAuthorizationView


class AuthorizationView(BaseAuthorizationView):
    """An authorization view that uses our own template."""

    template_name = 'oauth/authorize.html'



"""Unit tests for reviewboard.reviews.views.ReviewRequestViewMixin."""

from __future__ import unicode_literals

from datetime import datetime

import pytz
from django.utils import timezone

from reviewboard.reviews.models import ReviewRequest
from reviewboard.reviews.views import ReviewRequestViewMixin
from reviewboard.testing import TestCase


_local_timezone = pytz.timezone('US/Pacific')


class ReviewRequestViewMixinTests(TestCase):
    """Unit tests for reviewboard.reviews.views.ReviewRequestViewMixin."""

    fixtures = ['test_users']

    def test_get_review_request_status_html_with_submitted(self):
        """Testing ReviewRequestViewMixin.get_review_request_status_html
        with status=SUBMITTED
        """
        review_request = self.create_review_request(
            status=ReviewRequest.SUBMITTED,
            time_added=datetime(2018, 2, 10, 9, 23, 12, tzinfo=timezone.utc))

        mixin = ReviewRequestViewMixin()
        mixin.review_request = review_request

        with timezone.override(_local_timezone):
            html = mixin.get_review_request_status_html(
                review_request_details=review_request,
                close_info={
                    'timestamp': datetime(2018, 2, 12, 14, 56, 00,
                                          tzinfo=timezone.utc),
                })

        self.assertHTMLEqual(
            html,
            'Created Feb. 10, 2018 and submitted '
            '<time class="timesince" datetime="2018-02-12T06:56:00-08:00">'
            'Feb. 12, 2018, 6:56 a.m.</time>')

    def test_get_review_request_status_html_with_submitted_no_timestamp(self):
        """Testing ReviewRequestViewMixin.get_review_request_status_html
        with status=SUBMITTED and no timestamp
        """
        review_request = self.create_review_request(
            status=ReviewRequest.SUBMITTED,
            time_added=datetime(2018, 2, 10, 9, 23, 12, tzinfo=timezone.utc))

        mixin = ReviewRequestViewMixin()
        mixin.review_request = review_request

        with timezone.override(_local_timezone):
            html = mixin.get_review_request_status_html(
                review_request_details=review_request,
                close_info={
                    'timestamp': None,
                })

        self.assertEqual(html, 'Created Feb. 10, 2018 and submitted')

    def test_get_review_request_status_html_with_discarded(self):
        """Testing ReviewRequestViewMixin.get_review_request_status_html
        with status=DISCARDED
        """
        review_request = self.create_review_request(
            status=ReviewRequest.DISCARDED,
            time_added=datetime(2018, 2, 10, 9, 23, 12, tzinfo=timezone.utc))

        mixin = ReviewRequestViewMixin()
        mixin.review_request = review_request

        with timezone.override(_local_timezone):
            html = mixin.get_review_request_status_html(
                review_request_details=review_request,
                close_info={
                    'timestamp': datetime(2018, 2, 12, 14, 56, 00,
                                          tzinfo=timezone.utc),
                })

        self.assertHTMLEqual(
            html,
            'Created Feb. 10, 2018 and discarded '
            '<time class="timesince" datetime="2018-02-12T06:56:00-08:00">'
            'Feb. 12, 2018, 6:56 a.m.</time>')

    def test_get_review_request_status_html_with_discarded_no_timestamp(self):
        """Testing ReviewRequestViewMixin.get_review_request_status_html
        with status=DISCARDED and no timestamp
        """
        review_request = self.create_review_request(
            status=ReviewRequest.DISCARDED,
            time_added=datetime(2018, 2, 10, 9, 23, 12, tzinfo=timezone.utc))

        mixin = ReviewRequestViewMixin()
        mixin.review_request = review_request

        with timezone.override(_local_timezone):
            html = mixin.get_review_request_status_html(
                review_request_details=review_request,
                close_info={
                    'timestamp': None,
                })

        self.assertEqual(html, 'Created Feb. 10, 2018 and discarded')

    def test_get_review_request_status_html_with_pending_review(self):
        """Testing ReviewRequestViewMixin.get_review_request_status_html
        with status=PENDING_REVIEW
        """
        review_request = self.create_review_request(
            status=ReviewRequest.PENDING_REVIEW,
            time_added=datetime(2018, 2, 10, 9, 23, 12, tzinfo=timezone.utc),
            last_updated=datetime(2018, 2, 10, 15, 19, 23,
                                  tzinfo=timezone.utc))

        mixin = ReviewRequestViewMixin()
        mixin.review_request = review_request

        with timezone.override(_local_timezone):
            html = mixin.get_review_request_status_html(
                review_request_details=review_request,
                close_info={})

        self.assertHTMLEqual(
            html,
            'Created Feb. 10, 2018 and updated '
            '<time class="timesince" datetime="2018-02-10T07:19:23-08:00">'
            'Feb. 10, 2018, 7:19 a.m.</time>')

    def test_get_review_request_status_html_with_extra_info(self):
        """Testing ReviewRequestViewMixin.get_review_request_status_html
        with extra_info
        """
        review_request = self.create_review_request(
            status=ReviewRequest.PENDING_REVIEW,
            time_added=datetime(2018, 2, 10, 9, 23, 12, tzinfo=timezone.utc),
            last_updated=datetime(2018, 2, 10, 15, 19, 23,
                                  tzinfo=timezone.utc))

        mixin = ReviewRequestViewMixin()
        mixin.review_request = review_request

        with timezone.override(_local_timezone):
            html = mixin.get_review_request_status_html(
                review_request_details=review_request,
                close_info={},
                extra_info=[
                    {
                        'text': '{var} updated at {timestamp}',
                        'timestamp': datetime(2018, 2, 11, 23, 32, 00,
                                              tzinfo=timezone.utc),
                        'extra_vars': {
                            'var': 'Thingie',
                        },
                    },
                ])

        self.assertHTMLEqual(
            html,
            'Created Feb. 10, 2018 and updated '
            '<time class="timesince" datetime="2018-02-10T07:19:23-08:00">'
            'Feb. 10, 2018, 7:19 a.m.</time>'
            ' &mdash; '
            'Thingie updated at '
            '<time class="timesince" datetime="2018-02-11T15:32:00-08:00">'
            'Feb. 11, 2018, 3:32 p.m.</time>')



"""A backend for the Whoosh search engine."""

from __future__ import unicode_literals

import os

from django import forms
from django.conf import settings
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _

from reviewboard.search.search_backends.base import (SearchBackend,
                                                     SearchBackendForm)


class WhooshConfigForm(SearchBackendForm):
    """A form for configuring the Whoosh search backend."""

    search_index_file = forms.CharField(
        label=_("Search index directory"),
        help_text=_("The directory that search index data should be stored "
                    "in."),
        widget=forms.TextInput(attrs={'size': '80'}))

    def clean_search_index_file(self):
        """Clear the search_index_file field.

        This ensures the value is an absolute path and is writable.
        """
        index_file = self.cleaned_data['search_index_file'].strip()

        if index_file:
            if not os.path.isabs(index_file):
                raise ValidationError(
                    _("The search index path must be absolute."))

            if (os.path.exists(index_file) and
                    not os.access(index_file, os.W_OK)):
                raise ValidationError(
                    _('The search index path is not writable. Make sure the '
                      'web server has write access to it and its parent '
                      'directory.'))

        return index_file


class WhooshBackend(SearchBackend):
    """The Whoosh search backend."""

    search_backend_id = 'whoosh'
    name = _('Whoosh')
    haystack_backend_name = 'haystack.backends.whoosh_backend.WhooshEngine'
    config_form_class = WhooshConfigForm
    default_settings = {
        'PATH': os.path.join(settings.SITE_DATA_DIR, 'search-index'),
    }
    form_field_map = {
        'search_index_file': 'PATH',
    }



from __future__ import unicode_literals

import json
import logging

from django import template
from django.utils import six
from django.utils.html import mark_safe

from reviewboard.avatars import avatar_services


register = template.Library()


@register.simple_tag(takes_context=True)
def avatar(context, user, size, service_id=None):
    """Render the user's avatar to HTML.

    When the ``service_id`` argument is not provided, or the specified service
    is not registered or is not enabled, the user's specified avatar service
    will be used for rendering instead.

    Args:
        context (django.template.Context):
            The template rendering context.

        user (django.contrib.auth.models.User):
            The user whose avatar is to be rendered.

        size (int):
            The height and width of the avatar, in pixels.

        service_id (unicode, optional):
            The unique identifier of the avatar service to use. If this is
            omitted, or the specified service is not registered and enabled,
            the default avatar service will be used.

    Returns:
        django.utils.safestring.SafeText:
        The user's avatar rendered to HTML, or an empty string if no avatar
        service could be found.
    """
    service = avatar_services.for_user(user, service_id)

    if service is None:
        logging.error('Could not get a suitable avatar service for user %s.',
                      user)
        return mark_safe('')

    return service.render(request=context['request'], user=user, size=size)


@register.simple_tag(takes_context=True)
def avatar_url(context, user, size, resolution='1x', service_id=None):
    """Return the URL of the requested avatar.

    Args:
        context (django.template.Context):
            The template rendering context.

        user (django.contrib.auth.models.User):
            The user whose avatar is to be rendered.

        size (int):
            The height and width of the avatar, in pixels.

        resolution (unicode, optional):
            The resolution of the avatar. This should be one of ``'1x'``, for
            normal DPI, or ``'2x'``, for high DPI. This defaults to normal DPI.

        service_id (unicode, optional):
            The unique identifier of the avatar service to use. If this is
            omitted, or the specified service is not registered and enabled,
            the default avatar service will be used.

    Returns:
        django.utils.safestring.SafeText:
        The URL of the requested avatar, or an empty string if no avatar
        service could be found.
    """
    if resolution not in ('1x', '2x'):
        raise ValueError('resolution should be "1x" or "2x", not %r.'
                         % resolution)

    service = avatar_services.for_user(user, service_id)

    if service is None:
        logging.error('Could not get a suitable avatar service for user %s.',
                      user)
        return mark_safe('')

    urls = service.get_avatar_urls(request=context['request'],
                                   user=user,
                                   size=size)
    return urls[resolution]


@register.simple_tag(takes_context=True)
def avatar_urls(context, user, size, service_id=None):
    """Serialize the user's avatar URLs into a JavaScript object.

    Args:
        context (django.template.Context):
            The template rendering context.

        user (django.contrib.auth.models.User):
            The user whose avatar URLs are to be serialized.

        size (int):
            The height and width of the avatar, in pixels.

        service_id (unicode, optional):
            The unique identifier of the avatar service to use. If this is
            omitted, or the specified service is not registered and enabled,
            the default avatar service will be used.

    Returns:
        django.utils.safestring.SafeText:
        The rendered JavaScript object.
    """
    service = avatar_services.for_user(user, service_id)

    if service is None:
        logging.error('Could not get a suitable avatar service for user %s.',
                      user)
        urls = {}
    else:
        urls = {
            resolution: url
            for resolution, url in six.iteritems(
                service.get_avatar_urls(request=context['request'],
                                        user=user,
                                        size=size)
            )
        }

    return mark_safe(json.dumps(urls, sort_keys=True))



"""Django model administration for OAuth2 applications."""

from __future__ import unicode_literals

import importlib

from django.contrib import admin
from django.contrib.admin.options import IS_POPUP_VAR
from django.contrib.admin.util import flatten_fieldsets
from django.utils.translation import ugettext_lazy as _
from djblets.forms.fieldsets import filter_fieldsets

from reviewboard.oauth.forms import (ApplicationChangeForm,
                                     ApplicationCreationForm)
from reviewboard.oauth.models import Application


class ApplicationAdmin(admin.ModelAdmin):
    """The model admin for the OAuth application model.

    The default model admin provided by django-oauth-toolkit does not provide
    help text for the majority of the fields, so this admin uses a custom form
    which does provide the help text.
    """

    form = ApplicationChangeForm
    add_form = ApplicationCreationForm
    raw_id_fields = ('local_site',)

    fieldsets = (
        (_('General Settings'), {
            'fields': ('name',
                       'enabled',
                       'user',
                       'redirect_uris',),
        }),

        (_('Client Settings'), {
            'fields': ('client_id',
                       'client_secret',
                       'client_type'),
        }),

        (_('Authorization Settings'), {
            'fields': ('authorization_grant_type',
                       'skip_authorization',
                       'local_site',),
        }),

        (_('Internal State'), {
            'description': _(
                '<p>This is advanced state that should not be modified unless '
                'something is wrong.</p>'
            ),
            'fields': ('original_user',
                       'extra_data'),
            'classes': ('collapse',),
        }),
    )

    add_fieldsets = tuple(filter_fieldsets(
        form=add_form,
        fieldsets=fieldsets,
        exclude_collapsed=False,
    ))

    def get_fieldsets(self, request, obj=None):
        """Return the appropriate fieldset.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.

            obj (reviewboard.oauth.models.Application, optional):
                The application being edited, if it already exists.

        Returns:
            tuple:
            The fieldset for either changing an Application (i.e., when
            ``obj is not None``) or the fieldset for creating an Application.
        """
        if obj is None:
            return self.add_fieldsets

        return super(ApplicationAdmin, self).get_fieldsets(request, obj=obj)

    def get_form(self, request, obj=None, **kwargs):
        """Return the form class to use.

        This method mostly delegates to the superclass, but hints that we
        should use :py:attr:`add_form` (and its fields) when we are creating
        the Application.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.

            obj (reviewboard.oauth.models.Application, optional):
                The application being edited, if it exists.

        Returns:
            type:
            The form class to use.
        """
        if obj is None:
            kwargs = kwargs.copy()
            kwargs['form'] = self.add_form
            kwargs['fields'] = flatten_fieldsets(self.add_fieldsets)

        return super(ApplicationAdmin, self).get_form(request, obj=obj,
                                                      **kwargs)

    def response_add(self, request, obj, post_url_continue=None):
        """Return the response for the ``add_view`` stage.

        This method will redirect the user to the change form after creating
        the application. We do this because the ``client_secret`` and
        ``client_id`` fields are generated by saving the form and it is likely
        the user will want to view and/or copy them after creating this
        Application.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.

            obj (reviewboard.oauth.models.Application):
                The application that was created.

            post_url_continue (unicode, optional):
                The next URL to go to.

        Returns:
            django.http.HttpResponse:
            A response redirecting the user to the change form.
        """
        if ('_addanother' not in request.POST and
            IS_POPUP_VAR not in request.POST):
            request.POST['_continue'] = 1

        return super(ApplicationAdmin, self).response_add(
            request,
            obj,
            post_url_continue=post_url_continue,
        )


# Ensure that the oauth2_provider admin modules is loaded so that we can
# replace their admin registration with our own. If we do not do this, we can't
# guarantee that it will be registered before we try to unregister it during
# unit tests.
importlib.import_module('oauth2_provider.admin')
admin.site.unregister(Application)
admin.site.register(Application, ApplicationAdmin)



"""E-mail notification callbacks."""

from __future__ import unicode_literals

from django.utils import timezone
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.notifications.email.message import (
    prepare_password_changed_mail,
    prepare_reply_published_mail,
    prepare_review_published_mail,
    prepare_review_request_mail,
    prepare_user_registered_mail,
    prepare_webapi_token_mail)
from reviewboard.notifications.email.utils import send_email
from reviewboard.reviews.models import ReviewRequest


def _update_email_info(obj, message_id):
    """Update the e-mail message information on the object.

    The ``email_message_id`` and ``time_emailed`` fields of the model will be
    updated.

    Args:
        obj (reviewboard.reviews.models.review.Review or
             reviewboard.reviews.models.review_request.ReviewRequest):
            The object for whiche-mail information will be updated.

        message_id (unicode):
            The new e-mail message ID.
    """
    obj.email_message_id = message_id
    obj.time_emailed = timezone.now()
    obj.save(update_fields=('email_message_id', 'time_emailed'))


def send_password_changed_mail(user):
    """Send an e-mail when a user's password changes.

    The e-mail will only be sent when the ``mail_send_password_changed_mail``
    :py:class:`~djblets.siteconfig.models.SiteConfiguration` setting is
    enabled.

    Args:
        user (django.contrib.auth.model.User):
            The user whose password changed.
    """
    siteconfig = SiteConfiguration.objects.get_current()

    if siteconfig.get('mail_send_password_changed_mail'):
        send_email(prepare_password_changed_mail, user=user)


def send_reply_published_mail(user, reply, trivial, **kwargs):
    """Send e-mail when a review reply is published.

    Listens to the :py:data:`~reviewboard.reviews.signals.reply_published`
    signal and sends an e-mail if this type of notification is enabled (through
    ``mail_send_review_mail`` site configuration).

    Args:
        user (django.contrib.auth.models.User):
            The user who published the reply.

        reply (reviewboard.reviews.models.Review):
            The reply that was published.

        trivial (bool):
            Whether or not the reply was marked as trivial when it was
            published. If ``True``, the e-mail will not be sent.

        **kwargs (dict):
            Unused keyword arguments from the signal.
    """
    siteconfig = SiteConfiguration.objects.get_current()

    if trivial or not siteconfig.get('mail_send_review_mail'):
        return

    review_request = reply.review_request

    if not review_request.public:
        return

    review = reply.base_reply_to

    message, sent = send_email(prepare_reply_published_mail,
                               user=user,
                               reply=reply,
                               review=review,
                               review_request=review_request)

    if sent:
        _update_email_info(reply, message.message_id)


def send_review_published_mail(user, review, request, to_owner_only,
                               **kwargs):
    """Send e-mail when a review is published.

    Listens to the :py:data:`~reviewboard.reviews.signals.review_published`
    signal and sends e-mail if this type of notification is enabled (through
    the ``mail_send_review_mail`` site configuration setting).

    Args:
        user (django.contrib.auth.models.User):
            The user that published the review.

        review (reviewboard.reviews.models.review.Review):
            The review that was published.

        to_owner_only (bool):
            Whether or not the mail should only be sent to the review request
            submitter.

        request (django.http.HttpRequest):
            The HTTP request that triggered this e-mail.

        **kwargs (dict):
            Additional keyword arguments.
    """
    siteconfig = SiteConfiguration.objects.get_current()

    if not siteconfig.get('mail_send_review_mail'):
        return

    review_request = review.review_request

    if not review_request.public:
        return

    message, sent = send_email(prepare_review_published_mail,
                               user=user,
                               review=review,
                               review_request=review_request,
                               request=request,
                               to_owner_only=to_owner_only)

    if sent:
        _update_email_info(review, message.message_id)


def send_review_request_closed_mail(user, review_request, close_type,
                                    **kwargs):
    """Send e-mail when a review request is closed.

    Listens to the
    :py:data:`~reviewboard.reviews.signals.review_request_closed` signal and
    sends an e-mail if this type of notification is enabled (through the
    ``mail_send_review_close_mail`` site configuration setting).

    Args:
        user (django.contrib.auth.models.User):
            The user who closed the review request.

        review_request (reviewboard.reviews.models.review_request.ReviewRequest):
            The review request that was closed.

        close_type (unicode):
            How the review request was closed.

        **kwargs (dict):
            Unused keyword arguments from the signal.
    """
    siteconfig = SiteConfiguration.objects.get_current()

    if not (siteconfig.get('mail_send_review_close_mail') and
            review_request.public):
        return

    message, sent = send_email(prepare_review_request_mail,
                               user=user,
                               review_request=review_request,
                               close_type=close_type)

    if sent:
        _update_email_info(review_request, message.message_id)


def send_review_request_published_mail(user, review_request, trivial,
                                       changedesc, **kwargs):
    """Send e-mail when a review request is published.

    Listens to the
    :py:data:`~reviewboard.reviews.signals.review_request_published` signal and
    sends an e-mail if this type of notification is enabled (through the
    ``mail_send_review_mail`` site configuration setting).

    If the review request publish was marked as trivial, the e-mail will not be
    sent.

    Args:
        user (django.contrib.auth.models.User):
            The user who published the review request.

        review_request (reviewboard.reviews.models.review_request.ReviewRequest):
            The review request that was published.

        trivial (bool):
            Whether or not the publish was marked as trivial.

        changedesc (reviewboard.changedescs.models.ChangeDescription, optional):
            The change description associated with the publish, if any.

        **kwargs (dict):
            Ignored keyword arguments from the signal.
    """
    siteconfig = SiteConfiguration.objects.get_current()

    if trivial or not siteconfig.get('mail_send_review_mail'):
        return

    if (not review_request.public or
        review_request.status == ReviewRequest.DISCARDED):
        return

    message, sent = send_email(prepare_review_request_mail,
                               user=user,
                               review_request=review_request,
                               changedesc=changedesc)

    if sent:
        _update_email_info(review_request, message.message_id)


def send_user_registered_mail(user, **kwargs):
    """Send e-mail when a user is registered.

    Listens for new user registrations and sends a new user registration
    e-mail to administrators, if this type of notification is enabled (through
    ``mail_send_new_user_mail`` site configuration).

    Args:
        user (django.contrib.auth.models.User):
            The user that registered.

        **kwargs (dict):
            Unused keyword arguments from the signal.
    """
    siteconfig = SiteConfiguration.objects.get_current()

    if not siteconfig.get('mail_send_new_user_mail'):
        return

    send_email(prepare_user_registered_mail,
               user=user)


def send_webapi_token_created_mail(instance, auto_generated=False, **kwargs):
    """Send e-mail when an API token is created.

    Args:
        instance (reviewboard.webapi.models.WebAPIToken):
            The token that has been created.

        should_send_email (bool, optional):
            Whether or not an e-mail should be sent.

        **kwargs (dict):
            Unused keyword arguments provided by the signal.
    """
    if not auto_generated:
        send_email(prepare_webapi_token_mail,
                   webapi_token=instance,
                   op='created')


def send_webapi_token_updated_mail(instance, **kwargs):
    """Send e-mail when an API token is updated.

    Args:
        instance (reviewboard.webapi.models.WebAPIToken):
            The token that was updated.

        **kwargs (dict):
            Unused keyword arguments provided by the signal.
    """
    send_email(prepare_webapi_token_mail,
               webapi_token=instance,
               op='updated')


def send_webapi_token_deleted_mail(instance, **kwargs):
    """Send e-mail when an API token is deleted.

    Args:
        instance (reviewboard.webapi.models.WebAPIToken):
            The token that has been created or updated.

        **kwargs (dict):
            Unused keyword arguments provided by the signal.
    """
    send_email(prepare_webapi_token_mail,
               webapi_token=instance,
               op='deleted')



"""Unit tests for the Unfuddle hosting service."""

from __future__ import unicode_literals

from django.utils.six.moves.urllib.error import HTTPError

from reviewboard.hostingsvcs.errors import RepositoryError
from reviewboard.hostingsvcs.testing import HostingServiceTestCase
from reviewboard.scmtools.crypto_utils import (decrypt_password,
                                               encrypt_password)
from reviewboard.scmtools.errors import FileNotFoundError


class UnfuddleTests(HostingServiceTestCase):
    """Unit tests for the Unfuddle hosting service."""

    service_name = 'unfuddle'
    fixtures = ['test_scmtools']

    default_account_data = {
        'password': encrypt_password(HostingServiceTestCase.default_password),
    }

    default_repository_extra_data = {
        'unfuddle_account_domain': 'mydomain',
        'unfuddle_project_id': 1,
        'unfuddle_repo_id': 2,
        'unfuddle_repo_name': 'myrepo',
    }

    def test_service_support(self):
        """Testing Unfuddle service support capabilities"""
        self.assertTrue(self.service_class.supports_bug_trackers)
        self.assertTrue(self.service_class.supports_repositories)

    def test_get_repository_fields_with_git(self):
        """Testing Unfuddle.get_repository_fields for Git"""
        self.assertEqual(
            self.get_repository_fields(
                'Git',
                fields={
                    'unfuddle_account_domain': 'mydomain',
                    'unfuddle_project_id': 1,
                    'unfuddle_repo_name': 'myrepo',
                }
            ),
            {
                'path': 'git@mydomain.unfuddle.com:mydomain/myrepo.git',
                'mirror_path': ('https://mydomain.unfuddle.com/git/'
                                'mydomain_myrepo/'),
            })

    def test_get_repository_fields_with_subversion(self):
        """Testing Unfuddle.get_repository_fields for Subversion"""
        self.assertEqual(
            self.get_repository_fields(
                'Subversion',
                fields={
                    'unfuddle_account_domain': 'mydomain',
                    'unfuddle_project_id': 1,
                    'unfuddle_repo_name': 'myrepo',
                }
            ),
            {
                'path': 'https://mydomain.unfuddle.com/svn/mydomain_myrepo',
                'mirror_path': ('http://mydomain.unfuddle.com/svn/'
                                'mydomain_myrepo'),
            })

    def test_authorize(self):
        """Testing Unfuddle.authorize"""
        hosting_account = self.create_hosting_account(data={})

        with self.setup_http_test(payload=b'{}',
                                  hosting_account=hosting_account,
                                  expected_http_calls=1) as ctx:
            self.assertFalse(ctx.service.is_authorized())

            ctx.service.authorize(username='myuser',
                                  password='abc123',
                                  unfuddle_account_domain='mydomain')

        ctx.assertHTTPCall(
            0,
            url='https://mydomain.unfuddle.com/api/v1/account/',
            username='myuser',
            password='abc123',
            headers={
                'Accept': 'application/json',
            })

        self.assertIn('password', hosting_account.data)
        self.assertNotEqual(hosting_account.data['password'], 'abc123')
        self.assertEqual(decrypt_password(hosting_account.data['password']),
                         'abc123')
        self.assertTrue(ctx.service.is_authorized())

    def test_check_repository(self):
        """Testing Unfuddle.check_repository"""
        payload = self.dump_json([{
            'id': 2,
            'abbreviation': 'myrepo',
            'system': 'git',
        }])

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            ctx.service.check_repository(unfuddle_account_domain='mydomain',
                                         unfuddle_repo_name='myrepo',
                                         tool_name='Git')

        ctx.assertHTTPCall(
            0,
            url='https://mydomain.unfuddle.com/api/v1/repositories/',
            headers={
                'Accept': 'application/json',
            })

    def test_check_repository_with_wrong_repo_type(self):
        """Testing Unfuddle.check_repository with wrong repo type"""
        payload = self.dump_json([{
            'id': 2,
            'abbreviation': 'myrepo',
            'system': 'svn',
        }])

        expected_message = 'A repository with this name was not found'

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            with self.assertRaisesMessage(RepositoryError, expected_message):
                ctx.service.check_repository(
                    unfuddle_account_domain='mydomain',
                    unfuddle_repo_name='myrepo',
                    tool_name='Git')

        ctx.assertHTTPCall(
            0,
            url='https://mydomain.unfuddle.com/api/v1/repositories/',
            headers={
                'Accept': 'application/json',
            })

    def test_get_file_with_svn_and_base_commit_id(self):
        """Testing Unfuddle.get_file with Subversion and base commit ID"""
        self._test_get_file(
            tool_name='Subversion',
            revision='123',
            base_commit_id='456',
            expected_revision='456')

    def test_get_file_with_svn_and_revision(self):
        """Testing Unfuddle.get_file with Subversion and revision"""
        self._test_get_file(
            tool_name='Subversion',
            revision='123',
            base_commit_id=None,
            expected_revision='123')

    def test_get_file_with_git_and_base_commit_id(self):
        """Testing Unfuddle.get_file with Git and revision with base commit ID
        """
        self._test_get_file(
            tool_name='Git',
            revision='123',
            base_commit_id='456',
            expected_revision='456')

    def test_get_file_with_git_and_revision(self):
        """Testing Unfuddle.get_file with Git and revision without base commit
        ID
        """
        self._test_get_file(
            tool_name='Git',
            revision='123',
            base_commit_id=None,
            expected_revision=None,
            expected_error=True)

    def test_get_file_exists_with_svn_and_base_commit_id(self):
        """Testing Unfuddle.get_file_exists with Subversion and base commit ID
        """
        self._test_get_file_exists(
            tool_name='Subversion',
            revision='123',
            base_commit_id='456',
            expected_revision='456',
            expected_found=True)

    def test_get_file_exists_with_svn_and_revision(self):
        """Testing Unfuddle.get_file_exists with Subversion and revision"""
        self._test_get_file_exists(
            tool_name='Subversion',
            revision='123',
            base_commit_id=None,
            expected_revision='123',
            expected_found=True)

    def test_get_file_exists_with_svn_and_revision_not_found(self):
        """Testing Unfuddle.get_file_exists with Subversion and revision not
        found
        """
        self._test_get_file_exists(
            tool_name='Subversion',
            revision='123',
            base_commit_id=None,
            expected_revision='123',
            expected_found=False)

    def test_get_file_exists_with_git_and_base_commit_id(self):
        """Testing Unfuddle.get_file_exists with Git and revision with base
        commit ID
        """
        self._test_get_file_exists(
            tool_name='Git',
            revision='123',
            base_commit_id='456',
            expected_revision='456',
            expected_found=True)

    def test_get_file_exists_with_git_and_revision_no_base_commit_id(self):
        """Testing Unfuddle.get_file_exists with Git and revision without
        base commit ID
        """
        self._test_get_file_exists(
            tool_name='Git',
            revision='123',
            base_commit_id=None,
            expected_revision=None,
            expected_found=False,
            expected_error=True)

    def test_get_file_exists_with_git_and_revision_not_found(self):
        """Testing Unfuddle.get_file_exists with Git and revision not found"""
        self._test_get_file_exists(
            tool_name='Git',
            revision='123',
            base_commit_id='456',
            expected_revision='456',
            expected_found=False)

    def _test_get_file(self, tool_name, revision, base_commit_id,
                       expected_revision, expected_error=False):
        """Common function for testing file fetching.

        Args:
            tool_name (unicode):
                The registered name of the SCMTool.

            revision (unicode):
                The revision to fetch.

            base_commit_id (unicode):
                The ID the commit is based on.

            expected_revision (unicode):
                The expected revision to find in the URL.

            expected_error (bool, optional):
                Whether this test expects the file existence check to return
                an error.
        """
        if expected_error:
            expected_http_calls = 0
        else:
            expected_http_calls = 1

        with self.setup_http_test(expected_http_calls=expected_http_calls,
                                  payload=b'My data') as ctx:
            repository = ctx.create_repository(tool_name=tool_name)
            get_file_kwargs = {
                'repository': repository,
                'path': '/path',
                'revision': revision,
                'base_commit_id': base_commit_id,
            }

            if expected_error:
                with self.assertRaises(FileNotFoundError):
                    ctx.service.get_file(**get_file_kwargs)

                result = None
            else:
                result = ctx.service.get_file(**get_file_kwargs)

        if not expected_error:
            ctx.assertHTTPCall(
                0,
                url=('https://mydomain.unfuddle.com/api/v1/repositories/2/'
                     'download/?path=/path&commit=%s'
                     % expected_revision),
                headers={
                    'Accept': 'application/json',
                })

            self.assertIsInstance(result, bytes)
            self.assertEqual(result, b'My data')

    def _test_get_file_exists(self, tool_name, revision, base_commit_id,
                              expected_revision, expected_found=True,
                              expected_error=False):
        """Common function for testing file existence checks.

        Args:
            tool_name (unicode):
                The registered name of the SCMTool.

            revision (unicode):
                The revision to fetch.

            base_commit_id (unicode):
                The ID the commit is based on.

            expected_revision (unicode):
                The expected revision to find in the URL.

            expected_found (bool, optional):
                Whether this test expects the check to indicate the file
                exists.

            expected_error (bool, optional):
                Whether this test expects the file existence check to return
                an error.
        """
        if expected_found:
            payload = b'{}'
            status_code = None
        else:
            payload = None
            status_code = 404

        if expected_error:
            expected_http_calls = 0
        else:
            expected_http_calls = 1

        with self.setup_http_test(expected_http_calls=expected_http_calls,
                                  payload=payload,
                                  status_code=status_code) as ctx:
            repository = ctx.create_repository(tool_name=tool_name)
            result = ctx.service.get_file_exists(repository=repository,
                                                 path='/path',
                                                 revision=revision,
                                                 base_commit_id=base_commit_id)

        if expected_error:
            self.assertEqual(len(ctx.http_calls), 0)
            self.assertFalse(result)
        else:
            ctx.assertHTTPCall(
                0,
                url=('https://mydomain.unfuddle.com/api/v1/repositories/2/'
                     'history/?path=/path&commit=%s&count=0'
                     % expected_revision),
                headers={
                    'Accept': 'application/json',
                })

            if expected_found:
                self.assertTrue(result)
            else:
                self.assertTrue(ctx.http_calls[0].raised(HTTPError))
                self.assertFalse(result)
